/*************************************************************************/
/*                           PROPERTY OF THE                             */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/
#include "AQS.h"  
#include "time.h"
 
/*************************************************************************/
/*                     AQSTransition Message Definitions    	         */
/*************************************************************************/
/*************************************************************************/
/*  		        Default Constructor Definition   	         */
/*************************************************************************/

AQSTransition::AQSTransition()  { }
AQSTransition::~AQSTransition() { }   

/*************************************************************************/
/*  		          Assignent Operator Overloading                 */
/*************************************************************************/

AQSTransition& AQSTransition::operator=(AQSTransition &arg) 
{ if(this != &arg)
    { ProcessIndex  = arg.ProcessIndex;
      RepPtr        =       arg.RepPtr;
      Permutation   =  arg.Permutation;
    }
  return(*this);
}


int* AQSTransition::getActualPermutation()    { return(Permutation); }
int  AQSTransition::getProcessIndex()         { return(ProcessIndex); }
int  AQSTransition::getRepPtr()               { return(RepPtr); } 
void AQSTransition::setPermutation(int *Perm) { Permutation = Perm; }
void AQSTransition::setRepPtr(int pS)         { RepPtr = pS; }
void AQSTransition::setProcessIndex(int Indx) { ProcessIndex = Indx; }

/*************************************************************************/
/*        	        Display() Message Definition		         */
/*************************************************************************/

void AQSTransition::Display()
{
 cout<<"\tProcess: " << ProcessIndex;
 cout<<",\ttarget: " << RepPtr;
 cout<<",\tpermtn: ";
 for(int i=0; i<AQSState::NumberOfProcesses; i++) cout<<Permutation[i];
 cout << endl;
}

ostream& operator<<(ostream &Cout, AQSTransition &arg)
{ arg.Display(); return(Cout); }

/*************************************************************************/
/*                   AQSState Message Definitions                        */
/*************************************************************************/ 
/*************************************************************************/
/* 		       Default Constructor Definition          		 */
/*************************************************************************/

int    AQSState::TableSize;
int    AQSState::NumberOfProcesses;
char **AQSState::VTNames;
void FindPartition(int *, int *, Program& );

AQSState::AQSState(int *SValues_, long cs)
{ SValues           = SValues_;  
  Link              = -1; 
  partition         = new int[NumberOfProcesses]; 
  ExploredBit       = -1;
  BbarLink          = NULL;
  CheckSum          = cs;
}

AQSState::~AQSState() { }

void AQSState::InitializePartition(Program *prog)
{ if(ExploredBit != -1) return;
  ExploredBit = 0;

  FindPartition(SValues, partition, *prog);
  // This sets DISABLED bit for every process
  // In the next case we set it correctly (clear it for enabled)
  if(OnFlyOption == 2)
    { int            *Table        = new int[TableSize]; //Dummy	
      Transition     *trans;
      int            *PIndex; 
      
      for(trans  = prog->TransitionIteratorReset(); 
	  trans != NULL; 
	  trans  = prog->TransitionIteratorNext())
	{ PIndex = partition + trans->GetProcess()->GetIndexOfProcess();
	  if((*PIndex & 1) &&  // If it is a LEADER
	     (*PIndex & 2) &&  // and still DISABLED
	     (trans->TakeAction2(Table, SValues, TableSize))) // and EXECUTED
	    *PIndex -= 2;      // then clear the DISABLED bit   
	}    
      delete [] Table;
    }
}
  
/************************************************************************/
/* 	       Some get and set functions     Definition	      	*/
/************************************************************************/

int* AQSState::getActualValueTable() { return(SValues); }
int  AQSState::getLink()             { return(Link); }
void AQSState::setLink(int number)   { Link = number; }
int  AQSState::getValueTableSize()   { return(TableSize);}
int  AQSState::getTransitionsSize()  { return(Transitions.getSize());}
void AQSState::AddTransition(AQSTransition &arg)
                                     { Transitions.Append(arg); }
void AQSState::releaseTransitionsHead()
                                     { Transitions.Deallocate(); }
Node<AQSTransition>* AQSState::getTransitionsHead() 
                                     { return(Transitions.getHead()); }

/************************************************************************/
/*  		       Display() Message Definition	    	        */
/************************************************************************/
void AQSState::PrintValueTable()
{ for( int i=0; i<TableSize; ++i ) cout << ' ' 
    << *(VTNames[i])<< *(VTNames[i]+1) << SValues[i];
}


void AQSState::Display()
{ int i;

  cout << endl;
  PrintValueTable();
  cout << endl << "Partition:   ";
  for(i=0;i<NumberOfProcesses;i++) cout <<  (partition[i]>>2);
  cout <<  endl << "Statusbits:  ";
  for(i=0;i<NumberOfProcesses;i++) cout <<  (partition[i]&3);
  cout << endl << "It has " << Transitions.getSize() << " transitions: " << endl;
  Transitions.Display(); 
  cout<<endl<<endl;
}

ostream& operator<<(ostream &Cout, AQSState &arg)
{
   arg.Display(); 
   return(Cout);
}
 
/*************************************************************************/
/*                             AQS Message Definitions                   */
/*************************************************************************/ 
long CheckSum(int *a, Program* prog)
{ int res;
  long cs = 0;
  
  int i,j;
  Node<Variable*> *m;

  for(m = prog->GetVariablesHead(); m != NULL; m = m->Next)
   { j = m->data->size;
     res = 1;
     for(i = 0; i<j; i++) res += *(a++); 
     cs = (cs + res) * 11;    
   }
  return cs;
}

#define   StartingSize 500
#define   Offset       500

bool StateEquivalence(int*,int*,int*, Program&);

/*************************************************************************/
/*                         Default Constructor                           */
/*************************************************************************/

AQS::AQS(Program  &prg)
: map(prg.GetTableSize()), 
  PermutationStore(prg.GetNumberOfProcesses())
{ 
 long                 Q = 0;       // The first not explored AQS node
 int                 *Table;
 int                 *Perm;        // The array holding the permutation.

 AQSState::TableSize         = prg.GetTableSize();  
 AQSState::NumberOfProcesses = prg.GetNumberOfProcesses();  
 AQSState::VTNames           = prg.GetValueTableNames();

 prog      =   & prg; 
 NumberOfProcesses = prg.GetNumberOfProcesses();
 TotalSize =   StartingSize;
 Size      =   0;
 STable    =   new AQSState*[TotalSize];
 assert(STable != 0);
 
 Table       = new int[prg.GetTableSize()];	
 Perm        = new int[NumberOfProcesses]; 
 memcpy(Perm,PermutationStore.getIdPerm() , sizeof(int)*NumberOfProcesses);

 prg.GetTable(Table);
 assert(Insert(new AQSState(Table,CheckSum(Table,prog))) == 0);
      
 map[map.hash(Table)] = 0;      // First state inserted in the BO[0] position. 

 if(!OnFlyOption)
   { if(!(ExampleOption&4) && ExampleOption) cout <<"AQS States:"<<flush;
     while (Q < Size)  Explore(Q++);
   }
}


/***********************************************************************/
/*                         Default Destructor                          */
/***********************************************************************/

AQS::~AQS() 
{ 
  delete  []  STable; 
  STable    =   NULL;
  Size      =      0;
  TotalSize =      0;
}



/***********************************************************************/
/*                 the routine that explores an AQS node               */
/***********************************************************************/
                          
void AQS::Explore(long index)
{
 Transition          *trans;
 AQSState            *StartPtr;  
 int                 *StartTable;
 long                 tempValue;	
 int                  PIndex; 
 long                 i;
 long                 hashedIndex; // Hash table index returned after hashing.
 long                 lastHashedState; // Last BO vertice with Link = -1.
 int                  TableSize;
 int                 *Perm; // The array holding the permutation.	
 int                 *Table;
 AQSTransition        T;	
 long                 cs;

 assert(index < Size);
 StartPtr   = STable[index]; 
 if(StartPtr->ExploredBit == -1) StartPtr->InitializePartition(prog);
 if(StartPtr->ExploredBit++) return;


 TableSize   = prog->GetTableSize();
 Table       = new int[TableSize];	
 Perm        = new int[NumberOfProcesses]; 

 memcpy(Perm,PermutationStore.getIdPerm() , sizeof(int)*NumberOfProcesses);   
 StartTable = StartPtr->getActualValueTable(); 	    	

 for(trans  = prog->TransitionIteratorReset(); 
     trans != NULL; 
     trans  = prog->TransitionIteratorNext())
   { // get Process Identifier and Indeces from each ProcessBody.   
     PIndex = trans->GetProcess()->GetIndexOfProcess();
     if(StartPtr->partition[PIndex] & 1) 
       // This `if' reveals if the process is a leading process in the
       // equivalence class - the low bit indicates it in the partition.
       if(trans->TakeAction(Table, StartTable, TableSize))
	 { // There is an enabled transition, so this leading process is not
	   // DISABLED. The appropriate flag shoud be cleared. After the
	   // routine only those DISABLED flags remain high, that are really
	   // disabled. 
	   if(StartPtr->partition[PIndex] & 2) 
	     StartPtr->partition[PIndex] -= 2;
	   
	   T.setProcessIndex(PIndex);
	   cs = CheckSum(Table, prog);
	   for(i = map[hashedIndex = map.hash(Table)]; 
	       i != -1;
	       lastHashedState = i, i = STable[i]->getLink())
	     if(cs == STable[i]->CheckSum &&
		StateEquivalence(STable[i]->getActualValueTable(),
				 Table, Perm, *prog))
	       { T.setRepPtr(i);
		 T.setPermutation(PermutationStore.Insert(Perm));
		 break;
	       }
	   if(i == -1)
	     { tempValue = Insert(new AQSState(Table,cs));  
	       Table = new int[TableSize];	

	       if(!OnFlyOption && !(ExampleOption&4) && ExampleOption && 
		  (tempValue % 100) == 0 ) 
		 cout <<((tempValue % 1000)?".":"*")<<flush;
		 
	       if( map[hashedIndex] == -1 )      
		 map[hashedIndex] = tempValue;   // Un-occupied cell.  
	       else
		 STable[lastHashedState]->setLink(tempValue);	
	       
	       T.setRepPtr(tempValue);
	       T.setPermutation(PermutationStore.getIdPerm());   
	     }         	 
	   StartPtr->AddTransition(T);

	 }      // of if(TakeAction)
                // of if(StartPtr->partition[PIndex] & 1)...
   }            // of the for(trans) statement. 
 if(ExampleOption&4) 
   cout << endl << "AQS node " << index << " was created!" << *StartPtr;
 delete [] Perm;   
 delete [] Table;
 StartPtr->ExploredBit = 1;
}                               

/***********************************************************************/
/*                 Deadlock tester                                     */
/***********************************************************************/
                             
bool AQS::deadlock()            
{ 
 long                 index = 0;
 Transition          *trans;
 AQSState            *StartPtr;  
 int                 *StartTable;
 long                 tempValue;	
 int                  PIndex; 
 long                 i;
 long                 hashedIndex; // Hash table index returned after hashing.
 long                 lastHashedState; // Last BO vertice with Link = -1.
 int                  TableSize;
 int                 *Perm; // The array holding the permutation.	
 int                 *Table;
 bool                 deadlock;
 long                 cs;

 TableSize   = prog->GetTableSize();
 Table       = new int[TableSize];	
 Perm        = new int[NumberOfProcesses]; 

 while (index < Size)  
   { deadlock   = true;
     StartPtr   = STable[index]; 
     if(StartPtr->ExploredBit == -1) StartPtr->InitializePartition(prog);
     StartTable = StartPtr->getActualValueTable(); 	    	

     for(trans  = prog->TransitionIteratorReset(); 
	 trans != NULL; 
	 trans  = prog->TransitionIteratorNext())
       { PIndex = trans->GetProcess()->GetIndexOfProcess();
	 if((StartPtr->partition[PIndex] & 1) && 
	     trans->TakeAction(Table, StartTable, TableSize))
	   { deadlock = false;
	     cs = CheckSum(Table, prog);
	     for(i = map[hashedIndex = map.hash(Table)]; 
		 i != -1;
		 lastHashedState = i, i = STable[i]->getLink())
	     if(cs == STable[i]->CheckSum &&
		StateEquivalence(STable[i]->getActualValueTable(),
				   Table, Perm, *prog))
		 break;	       
	     if(i == -1)
	       { tempValue = Insert(new AQSState(Table,cs));  
		 Table = new int[TableSize];	
		 assert(Table != NULL);

		 if(!(ExampleOption&4) && ExampleOption && 
		    (tempValue % 100) == 0 ) 
		   cout <<((tempValue % 1000)?".":"*")<<flush;
		 
		 if(map[hashedIndex] == -1) map[hashedIndex] = tempValue;
		 else STable[lastHashedState]->setLink(tempValue);	
	       }         	 
	   }   
       }           
     if(ExampleOption&4) 
       cout << endl << "AQS node " << index << " was created!" << *StartPtr;
     if(deadlock) return true;
    index++;
   }
 delete [] Perm;   
 delete [] Table;
 return false;
}                               



/***********************************************************************/
/*                        Grow() Message Definition                    */
/*     Grows the receiver's structure when it becomes 90% occupied.    */
/***********************************************************************/


void AQS::Grow() 
{ 
  AQSState   **oldAddr;
      
  oldAddr = STable;
  TotalSize += Offset;
  STable = new AQSState*[TotalSize];
  
  if(STable == NULL)
    {
      cout<<endl;
      cout<<"**ERROR** In AQS::Grow()-memory allocation failed."<<endl;
      exit(1);
    }    
  
  memcpy(STable,oldAddr,sizeof(AQSState*)*Size);

  if(STable == oldAddr)   // Then reallocation failed.
    { 
      cout<<endl<<endl; 
      cout<<"**AQS ERROR** Reallocation Failed. Crash likely.";
      cout<<endl<<endl;
      exit(1);
    }
  delete [] oldAddr;
}

/***********************************************************************/
/*                     Insert() Message Definition                     */
/*   Inserts the parameter AQSState Object in the Base Object.         */
/***********************************************************************/

long AQS::Insert(AQSState *rhs)
{      
  if(Size == TotalSize)  Grow();
  STable[Size] = rhs; 
  
  return(Size++);             
}

/************************************************************************/
/*                      some basic functions like get                   */
/************************************************************************/

long  AQS::getSize()const { return(Size); }
AQSState** AQS::getHead() { return(STable); } 
AQSState&  AQS::operator[](long index) 
                { assert(index <= Size); return(*STable[index]);}

/************************************************************************/
/*                      the on-the-fly supplier of new AQS nodes        */
/************************************************************************/

Node<AQSTransition>* AQS::getTransitionsHead(long index)
{ Explore(index);
  return STable[index]->getTransitionsHead();
}

void AQS::releaseTransitionsHead(long index)
{ if(OnFlyOption<2) return;
  assert(index <= Size);
  assert(STable[index]->ExploredBit--);
  if(!STable[index]->ExploredBit) STable[index]->releaseTransitionsHead();
}

/************************************************************************/
/*                        Output Functions                              */
/************************************************************************/ 

ostream& operator<<(ostream &Cout, AQS &arg) 
{ arg.Display();  return(Cout); }

void AQS::Display()const 
{ for(int i=0; i<Size; i++) STable[i]->Display(); }

int AQS::Report() 
{ 
  int sum1,sum2,i;
  int transCounter = 0; 
  int psi = 0; 
  for(i=0; i<Size; i++) 
    transCounter += STable[i]->getTransitionsSize(); 
  cout<<"Number of AQS states ----------------------> "
    <<Size<<endl;
  cout<<"Size of an AQS state (bytes)---------------> "
    << sizeof(AQSState)<< " + " 
    <<AQSState::TableSize*sizeof(int)<< " + "
    <<sizeof(int)*NumberOfProcesses << endl;
  cout<<"Number of transitions in all AQS states ---> "
    <<transCounter<<endl;
  cout<<"Size of a Transition (bytes)---------------> "
    <<sizeof(AQSTransition)<<endl;
  sum2=(((sizeof(AQSState) + (sizeof(int) * AQSState::TableSize)
	                        + (sizeof(int) * NumberOfProcesses )) 
                            * Size) +
       ( sizeof(AQSTransition) * transCounter) +
       ( sizeof(int) * psi)
       )/1000;
  cout<<"Total size of AQS structure (kbytes) =============> "
    <<sum2<<endl; 

  cout<<"Number of permutations in the P.S. -------> "
    <<PermutationStore.getSize()<<endl;
  cout<<"Max number of perm's under a hash slot ---> "
    <<PermutationStore.getmaxdistr()<<endl;
      
  sum1 = ((sizeof(Leaf)+sizeof(int)*NumberOfProcesses)*
	  PermutationStore.getSize())/1000;
  sum2+= sum1;
  cout<<"Total size of the P.S. (kbytes) ==================> "
    <<sum1<<endl;
      

  return(sum2);
}  
/************************************************************************/
/*          Concurrent Software Systems Laboratory (c)1995              */
/************************************************************************/
