/*************************************************************************/
/*                           PROPERTY OF THE                             */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/
/*************************************************************************/
/*                             Header File                               */
/*************************************************************************/ 

#include      <malloc.h>
#include      "HashTable.h"
#include      "Iterator.h"
#include      "Trie.h"
#include      "Interface.h"

#ifndef TRANSITION_H_
#define TRANSITION_H_

extern int          ExampleOption;
extern int          OnFlyOption;

class Bbar;
class AQSState;   // Necessary Declaration.
  
class AQSTransition
{
 public:
  
     AQSTransition();  
    ~AQSTransition();  
     AQSTransition&   operator=(AQSTransition&);  
     int*     getActualPermutation();  
     int      getProcessIndex();  
     int      getRepPtr();  
     void     setPermutation(int*);  
     void     setRepPtr(int);  // Sets the state ID of the Equiv. state.
     void     setProcessIndex(int);  // Sets the process' Index.

     void     Display();  // Displays the AQSTransition object. 
     friend ostream&     operator<<(ostream&, AQSTransition&);

   private:

     int	  ProcessIndex; // process which caused the transition.
     int 	 *Permutation;  // The permutation.
     int          RepPtr;       // Shows the Equivalent state of destination
};
 

class AQSState
{ 
  public:
	 AQSState(int*,long);  
	~AQSState();  
   int*   getActualValueTable();  
   int    getValueTableSize();  
   int    getTransitionsSize();  
   int    getLink(); 
   void   setLink(int);
   void	  Display();  
   void	  PrintValueTable();   
   void   InitializePartition(Program *);
   void   AddTransition(AQSTransition&);  // Adds a transition on the Trans
   Node<AQSTransition>*  getTransitionsHead();  // To be used by the Iterator.
   void                  releaseTransitionsHead();
   static   int          NumberOfProcesses; 
   static   char       **VTNames;

   friend ostream& operator<<(ostream&, AQSState&);
   friend class AQS;
   friend class Bbar;
 
  private:
   Bbar                 *BbarLink;     // Pointer to a Bar state hashed here
   int                   ExploredBit;  // Are the transitions explored?
   int                  *SValues;      // Array of values 
   long                  CheckSum;     // A symmetry preserving hashvalue
   static   int          TableSize;  
   int                   Link;         // Links equivalent states.
   List<AQSTransition>   Transitions;  // List of transitions.
   int                  *partition;    // the initial partition
  };

class AQS
{  
  friend ostream& operator<<(ostream&, AQS&); 
  
 public: 
  
      	               AQS(Program&);     // Default Constructor.
       	              ~AQS();             // Default Destructor.
             
  AQSState&            operator[](long);  // Returns the object
  AQSState**           getHead();         // Returns the array's starting addr.
  long                 Insert(AQSState*);  
  long                 getSize()const;    // Returns the size of the array.
  void                 Display()const;    // Display the Base Object.
  int                  Report();
  void                 Grow();            // Grows the structure.   
  Trie                 PermutationStore;
  Node<AQSTransition>* getTransitionsHead(long);  
  void                 releaseTransitionsHead(long);
  void                 Explore(long);     // Explores the successors
  bool                 deadlock();        // tests for deadlock.

private: 
  Program             *prog;
  int                  NumberOfProcesses; 
  AQSState           **STable;            // Beggining of the array.
  long                 Size;              // Running Size of the table.
  long                 TotalSize;         // Totally allocated space.
  HashTable            map;  
};


#endif
/*************************************************************************/
/*            Concurrent Software Systems Laboratory (c) 1995            */
/*************************************************************************/






