/********************************************************************/
/*                         PROPERTY OF THE                          */
/*             CONCURRENT SOFTWARE SYSTEMS LABORATORY               */
/********************************************************************/
#include       "BBar.h"
#define MIN(a,b) (a<b?a:b)
/*******************************************************************/
/*                  Bbar Class Message Definition                  */
/*******************************************************************/
/*******************************************************************/
/*                INITIALIZATION OF STATIC VARIABLES               */
/*******************************************************************/

AQS*        Bbar::aqs           = NULL;
Program*    Bbar::prog          = NULL;
int         Bbar::LastMyName    =    0;
int         Bbar::Bbarcount     =    0;
clock_t     Bbar::BeginTime     =    0; 
int         Bbar::NumberOfProcesses; 
/*******************************************************************/
/*                 BBAR DESTRUCTOR DEFINITIONS                     */
/*         Deletes all the transitions from a Bbar state.          */
/*******************************************************************/

Bbar::~Bbar()
{
 
  Permutations = NULL; 
 
  DFnum          =    0;
  LowLink        =    0;
  status         =    0;
}

/*****************************************************************/
/*                     Constructor Definition			 */  
/*         External functions should call this one only.         */
/*****************************************************************/

Bbar::Bbar(AQS *aqs_, Program *Q)
{
 
  ThisAQSState   = 0; 
  automata_index = Q->GetFirstAutoState();
  Permutations   = aqs_->PermutationStore.getIdPerm();
  Link           = NULL;
  status         = 0;
  DFnum          = LastMyName;
  LowLink        = LastMyName++;
  NumberOfProcesses = Q->GetNumberOfProcesses();
  Bbarcount++;

  aqs            = aqs_;
  prog           = Q;
  BeginTime      = clock();  

  MarkVisited(); 
  MarkSingle();           
  MarkOnStack();

  partition      = new int[NumberOfProcesses];            
  InitializePartition();

  if(automata_index->isFinal()) MarkInFinal();
                                  else  MarkOutFinal();
  (*aqs)[ThisAQSState].BbarLink = this;
  if(ExampleOption&16) cout<<*this;
  else if(ExampleOption) cout << "BBar States:" << flush;
}

/*****************************************************************/
/*                    Constructor Definition                     */
/*        Recursive constructor from TraverseTransitions.        */
/*****************************************************************/

Bbar::Bbar(long equiv, AutoState *next_auto_state_,int *perm)
{ ThisAQSState   = equiv; 
  automata_index = next_auto_state_;
  Permutations   = aqs->PermutationStore.Insert(perm);
  Link           = (*aqs)[ThisAQSState].BbarLink ;
  status         = 0;
  DFnum          = LastMyName;
  LowLink        = LastMyName++;
  MarkVisited(); 
  MarkSingle();           
  MarkOnStack();
  Bbarcount++;

  partition      = new int[NumberOfProcesses];       
  InitializePartition();

  if(automata_index->isFinal()) MarkInFinal();
                          else  MarkOutFinal();

  (*aqs)[ThisAQSState].BbarLink = this; 
  
  if(ExampleOption&16) cout<<*this;
  else if(ExampleOption && (Bbarcount % 100) == 0 ) 
    cout <<((Bbarcount % 1000)?".":"*")<<flush;
}

/************************************************************************/
/*               Revitalizing a BBar node: WakeUp                       */
/************************************************************************/
void Bbar::WakeUp(int *perm)
{ Permutations   = aqs->PermutationStore.Insert(perm);
  WakeUp();
}

void Bbar::WakeUp()
{ InitializePartition();
  status         = 0;
  DFnum          = LastMyName;
  LowLink        = LastMyName++;
  MarkVisited(); 
  MarkSingle();           
  MarkOnStack();
  if(ExampleOption&16) cout << endl <<"The following BBar node is revitalized"
   << *this;
 
}

/************************************************************************/
/*               Partition Initialization and Deletion                  */
/************************************************************************/

void Bbar::InitializePartition()
{ 
  int i,j,k;
  int *ip;
  int *aqspart;
  int flag;
  
  if(OnFlyOption == 1)  aqs->Explore(ThisAQSState);
  if(OnFlyOption == 2) (*aqs)[ThisAQSState].InitializePartition(prog);
  aqspart        = (*aqs)[ThisAQSState].partition; 
  partition_size = NumberOfProcesses;
  ip             = new int[NumberOfProcesses];

  for( i = 0; i < NumberOfProcesses; i++ )	 
    aqspart[i] |= (aqspart[aqspart[i]>>2] & 2);

  for( i = 0; i < NumberOfProcesses; i++ )	 
    ip[Permutations[i]] =  aqspart[i]>>1;
   

/*  for( i = 0; i < NumberOfProcesses; i++ )        
    if(ip[i]&1)                                 // if process i is disabled
      for( j = 0; j < NumberOfProcesses; j++ )  // then set all fellow proc's
	if(ip[i]>>1 == ip[j]>>1) ip[j] |= 1;    // DISABLED bit
*/

  for(Node<TempVar*> *mp = prog->tracked.getHead();mp != NULL;mp = mp->Next)
    { i = mp->data->GetIndexOfProcess();
      ip[i] = ((partition_size++)<<1) + (ip[i]&1);
    }

  partition_size=0;
  for( i = 0; i < NumberOfProcesses; i++ )	 
    { flag = 1;
      for(j = 0; j<i; j++)
	if(ip[j]==ip[i])
	  { partition[i] = partition[j] - 1;              // clear LEADER bit
            flag=0; 
	    break;
	  }
      if(flag) partition[i] = ((partition_size++) << 3) +  // the class id
	                         ((1 - (ip[i]&1)) << 2) +  // ENABLED bit
          (((ip[i]&1) && (FairnessOption != 's')) << 1) +  // EXECUTE bit
	                                       (1 << 0);   // LEADER bit
    }

  delete [] ip;
}

void Bbar::FreePartition() { delete [] partition; }

/*****************************************************************/
/*             GetNewPermutations() Message Definition	         */
/*****************************************************************/

void Bbar::GetNewPermutations(int *new_perm, 
                              AQSTransition *trans, int m1, int m2)
{ int *Perm;
  int j;
  

    // We compute the product: Permutation x [m1,m2] x Perm
  Perm  = trans->getActualPermutation();
  for(j=0; j<NumberOfProcesses; ++j)
    new_perm[j]=(Perm[j] == m1 ? Permutations[m2] : 
		 (Perm[j] == m2 ? Permutations[m1] : Permutations[Perm[j]]));
}
  

/*********************************************************************/
/*             CheckRepetition() Message Definition 	   	     */
/*             PermutationsMatch() routine Definition                */
/*  If the equivalent node already exists, return a pointer to it.   */
/*  else return NULL.                                                */
/*********************************************************************/
Bbar* Bbar::CheckRepetition(long old_state,
                            AutoState *auto_state_index, int *Index)
{
  Bbar               *pNext;
        
  pNext = (*aqs)[old_state].BbarLink;    
       
  while(pNext != NULL)
    {    
      if ((old_state==pNext->ThisAQSState) &&
	  (auto_state_index==pNext->automata_index)&&
	  (pNext->PermutationsMatch(Index)))
	return(pNext);
      pNext = pNext->Link; 
    }
  return((Bbar*)NULL);
}

bool Bbar::PermutationsMatch(int *perm)
{
  int i,  j;
  int k1, k2;
  int    *aqspart;
  static int *newperm;
  static int  flag = 1;

  if(flag) { newperm = new int[NumberOfProcesses]; flag--;}

  aqspart = (*aqs)[ThisAQSState].partition;
  memcpy(newperm, perm, sizeof(int) * NumberOfProcesses);  

  for(Node<TempVar*> *mp = prog->tracked.getHead();mp != NULL;mp = mp->Next)
    { i = mp->data->GetIndexOfProcess();
      for(j=0;j<NumberOfProcesses;++j)
	{ if(newperm[j]      == i) k1 = j;
	  if(Permutations[j] == i) k2 = j;
	}
      if(aqspart[k1]>>2 != aqspart[k2]>>2) return(false);
      newperm[k1] = newperm[k2];
      newperm[k2] = i;
    }

  memcpy(perm, newperm, sizeof(int) * NumberOfProcesses);  
  return(true);
}


/*****************************************************************/
/*          Some status modifying functions                      */
/*****************************************************************/

void Bbar::MarkVisited()    { status = status | 1; }  
void Bbar::MarkUnVisited()  { status = status & (63-1); }
bool Bbar::isVisited()      { return(status & 1  ? true:false); }

void Bbar::MarkOnStack()    { status = status | 2; }
void Bbar::MarkOffStack()   { status = status & (63-2); }
bool Bbar::isOnStack()      { return(status & 2  ? true:false); }
 
void Bbar::MarkDeleted()    { status = status | 4; } 
bool Bbar::isDeleted()      { return(status & 4  ? true:false); }

void Bbar::MarkInFinal()    { status = status | 8; }
void Bbar::MarkOutFinal()   { status = status & (63-8); }
bool Bbar::isFinal()        { return(status & 8  ? true:false); }

void Bbar::MarkNotSingle()  { status = status | 16; }
void Bbar::MarkSingle()     { status = status & (63-16); }
bool Bbar::isNotSingle()    { return(status & 16 ? true:false); }

/************************************************************************/
/*               isSatisfactory()                                       */
/*               Determines if every process is satisfied               */
/************************************************************************/

int*  Bbar::isSatisfactory() 
{ static bool flag = true;
  static int *comp;
  static int *bad_ones;
  // return = NULL means true
  // return = bad_ones means false  

  if(flag) 
    { comp     =  new int[NumberOfProcesses]; 
      bad_ones =  new int[NumberOfProcesses]; 
      flag     =  false; 
    }
  
  if(FairnessOption == 's' && LowLink != DFnum) return(bad_ones);  
            // in strong fairness case we do not allow early termination

  if(!isFinal() || !isNotSingle()) return(bad_ones);            
            // if we have not found a final automaton state or 
            // we are in a state without outgoing egde then we are not
            // satisfied

  int i;
  int *result = NULL;

  for(i=0;i<partition_size;i++) comp[i]=0;

  for(i=0;i<NumberOfProcesses;i++)              // collecting data
    comp[partition[i]>>3] |= partition[i] & 6;  // OR of all ENABLE/EXEC bits

  for(i=0;i<NumberOfProcesses;i++)         
    { bad_ones[i] = comp[partition[i]>>3];      // for all process classes 
      if(bad_ones[i] == 4) result = bad_ones;   // ENABLED bit is high 
    }                                           // EXECUTE bit is low but
                                        
  return(result);
}

bool Bbar::Survive(int* bad_ones)
{ int *aqspart = (*aqs)[ThisAQSState].partition; 

  if(FairnessOption != 's') return false;

  // If the local process i is enabled (the DISABLED bit is low in the
  // AQS) but not executed (the global correspondance of i is listed
  // as a bad_one) then we return false.

  for(int i=0; i<NumberOfProcesses; i++)     
    if(bad_ones[Permutations[i]] == 4 && !(aqspart[i] & 2)) return false;
  return true;
}

 
/************************************************************************/
/*               MergePartition(int*, int*)                             */
/*               Updates the partition in the Bbar node                 */
/************************************************************************/

int m_status = 1;
int        *m_st;
int    *m_result;
int      *m_comp; 
int       **m_ll;
clock_t                     bt,et;

        
void srch(int  j, int  c)
{  static int k;
   static int *pt;

  m_result[j]  = c;
  m_comp[c]   |= m_st[j] ;
  while(m_ll[j][0])
	{ pt = m_ll[j];
          k = pt[pt[0]];
	  pt[0]--;
	  if(!m_result[k]) srch(k,c);
	}
}

void Bbar::MergePartitions(int *pm1, int *pm2)
{ int      i,a,b;
  int      c = 1;
  int change = 0;

  for(i=NumberOfProcesses-1;!change && i;i--)
    if(partition[pm2[i]] >> 3 != partition[pm1[i]] >> 3) change = 1;
  if(!change) return;
 
  if(m_status)
    {
      m_ll     =  new       int*[2*NumberOfProcesses];
      for(a=0;a<2*NumberOfProcesses;a++) 
	m_ll[a]=  new       int[NumberOfProcesses+1];
      m_st     =  new       int[2*NumberOfProcesses]; 
      m_result =  new       int[2*NumberOfProcesses]; 
      m_comp   =  new       int[NumberOfProcesses+1]; 
      m_status = 0;
    }

  for(i=0;i<partition_size;i++)
    { m_result[i] = 0;
      m_comp[i+1] = 0;
      m_st[i]     = 0;
      m_ll[i][0]  = 0;
    }

  for(i=0;i<NumberOfProcesses;i++)
    { 
      a = partition[pm2[i]] >> 3;      
      b = partition[pm1[i]] >> 3;
      if(a!=b)
	{ m_ll[a][0]++; m_ll[a][m_ll[a][0]] = b;
	  m_ll[b][0]++; m_ll[b][m_ll[b][0]] = a;
	}
      m_st[ partition[i] >> 3 ] |= (partition[i] & 6);
    }

  for(i=0;i<partition_size;i++) if(!m_result[i]) srch(i,c++);
 
  for(i=0;i<NumberOfProcesses;i++) partition[i] = 
    ((m_result[partition[i]>>3] - 1) << 3) + // new index of the 
    m_comp[m_result[partition[i]>>3]] +      // ENABL/EXEC bits - already shifted
    (partition[i] & 1);                      // original LEADER bit   

  partition_size = c-1;
}


void Bbar::MergePartitions(int *part, int size)
{ int      i,a,b;
  int      c = 1;

  if(m_status)
    {
      m_ll     =  new      int*[2*NumberOfProcesses];
      for(a=0;a<2*NumberOfProcesses;a++) m_ll[a] = new int[NumberOfProcesses+1];
      m_st     =  new       int[2*NumberOfProcesses]; 
      m_result =  new       int[2*NumberOfProcesses]; 
      m_comp   =  new       int[NumberOfProcesses+1]; 
      m_status = 0; 
    }


  for(i=0;i<partition_size+size;i++)
    { m_result[i] = 0; 
      m_st[i]     = 0;
      m_ll[i][0]  = 0;
    }

  for(i=0;i<partition_size;i++) m_comp[i+1]=0;

  for(i=0;i<NumberOfProcesses;i++)
    { 
      a = partition[i] >> 3;             
      b = partition_size + (part[i] >> 3);
      m_ll[a][0]++; m_ll[a][m_ll[a][0]] = b;
      m_ll[b][0]++; m_ll[b][m_ll[b][0]] = a;
      m_st[a] |= partition[i] & 6;
      m_st[b] |= part[i] & 6;
    }

  for(i=0;i<partition_size;i++) if(!m_result[i]) srch(i,c++);

  for(i=0;i<NumberOfProcesses;i++) partition[i] =
    ((m_result[partition[i]>>3] - 1) << 3) + // new index of the 
    m_comp[m_result[partition[i]>>3]] +      // ENABL/EXEC bits - already shifted
    (partition[i] & 1);                      // original LEADER bit   

  partition_size = c-1;
}

 
/************************************************************************/
/*               Search Utility for weak fairness                       */
/*               Recursively creates the complete Bbar Structure        */
/************************************************************************/

bool Bbar::Search()
{ 
  static List<Bbar*>             S;    
  static int                    *NewPermutations;
  static int                     flag  = 1; 
  static int                     displ = 0;   
  static Bbar                   *pB2;
  AQSTransition                  currentAQSTransition;
  Iterator<AQSTransition>        ITTrans;
  int                            enabled_proc;
  int                            leader_proc;
  Bbar                          *pB;
  bool                           fairFlag = true;
  Node<AutoTransition*>         *mp;
  AutoTransition                *auto_transition;
  Node<AQSTransition>           *NextNode;
  int                           *SatTest;
  bool                           survFlag;
  List<Bbar*>                    S2;    



  if(flag) 
    { NewPermutations = new int[NumberOfProcesses];
      flag = 0;
    }       

  if(ExampleOption & 24) displ = 1;

  S.Insert(pB = this); 
      
  ITTrans.Adjust(aqs->getTransitionsHead(ThisAQSState));

  for(mp = prog->a_transitions.getHead();
      mp != NULL;
      mp = mp->Next)
    { auto_transition  = mp->data;
      if(automata_index == auto_transition->departure &&
	 auto_transition->condition->Evaluate2(
         (*aqs)[ThisAQSState].getActualValueTable(), Permutations))
      { for(NextNode  = ITTrans.Reset();
	    NextNode != NULL;
	    NextNode  = ITTrans.Next())
	  { currentAQSTransition = NextNode->getData();
	    enabled_proc = currentAQSTransition.getProcessIndex();
	    for(leader_proc = 0; leader_proc<NumberOfProcesses; leader_proc++)
	      if((partition[Permutations[leader_proc]]&1) &&
		 (((*aqs)[ThisAQSState].partition[leader_proc])>>2 == enabled_proc))
		{ GetNewPermutations( NewPermutations, 
				     &currentAQSTransition,
				      enabled_proc, leader_proc);
		  pB = CheckRepetition(currentAQSTransition.getRepPtr(),
				       auto_transition->destination,
				       NewPermutations);
		  if(pB != NULL && pB->isVisited())
		    /***********************************/
		    /* The state is not a new one      */ 
		    /***********************************/
		    { if(pB->isOnStack())
			{ MarkNotSingle();
			  LowLink = MIN(LowLink, pB->LowLink);
			  if(fairFlag) 
			    partition[Permutations[leader_proc]] |= 2;
			  MergePartitions(pB->Permutations,NewPermutations);
			  if(pB->isFinal()) MarkInFinal();
			}
		    }
		  else
		    /***********************************/
		    /*  Transition to a new state      */
		    /***********************************/
		    { if(ExampleOption&16) 
			cout<<endl<<endl<<"Transition from "
			  << DFnum <<" to the following new BBar state."<<endl;
		      
		      if(pB != NULL) pB->WakeUp(NewPermutations);
		      else pB = new Bbar(currentAQSTransition.getRepPtr(),
					 auto_transition->destination,
					 NewPermutations);
		      
		      if(pB->Search())
			/***********************************/
			/*  The new state returned YES     */
			/***********************************/
			{ LowLink = MIN(LowLink, pB->LowLink);
			  pB2 = pB;
			  
			  if(!S.isEmpty() && pB2->LowLink == pB2->DFnum)
			    { if(displ) cout << endl << endl <<
				" The strongly connected component:" << endl;
			      do { pB = S.RemoveHead(); 
				   if(displ) cout << *pB;
				   pB->FreePartition();
				 } while(pB != pB2 && !S.isEmpty());
			      
			      if(displ && (FairnessOption != 's')) 
				cout << endl << endl <<
				     " The path leading to the scc:" << endl;
			      while(!S.isEmpty()) 
				{ pB = S.RemoveHead(); 
				  if(displ && (FairnessOption != 's')) 
				    cout << *pB;
				  pB->FreePartition();
				}
			      displ = 0;
			    }  
			  if(!DFnum)
			    { if(!S.isEmpty())
				{ if(displ) cout << endl << endl <<
				    " The strongly connected component:" 
				      << endl;
				  do { pB = S.RemoveHead(); 
				       if(displ) cout << *pB;
				       pB->FreePartition();
				     } while(pB != this);
				}
			      aqs->releaseTransitionsHead(ThisAQSState);
			    }
			  aqs->releaseTransitionsHead(pB2->ThisAQSState);
			  return(true);
			}
		      /***********************************/
		      /*  The new state returned NOT YET */
		      /***********************************/
		      if(pB->isOnStack())
			/***********************************/
			/*  But at least it is on stack    */
			/***********************************/
			{ MarkNotSingle();
			  LowLink = MIN(LowLink, pB->LowLink);
			  if(fairFlag) 
			    partition[Permutations[leader_proc]] |= 2;
			  
			  MergePartitions(pB->partition,pB->partition_size);
			  
			  if(pB->isFinal()) MarkInFinal();
			  if ((!fairFlag) && (isSatisfactory() == NULL)) 
			    return(true); 
			}    // if(pB->isOnStack())
		    }        // if(pB->isOnStack()) else
		}            // if((partition[Permutations[...
	  }                  // for(NextNode  = ...
	if(fairFlag) fairFlag = false;
	if(isSatisfactory() == NULL) return(true); 
      }
    }
  if(LowLink == DFnum)
    { SatTest = isSatisfactory();
      if(SatTest == NULL) return(true);
      survFlag = (bool)(isFinal() && isNotSingle());
      do { pB = S.RemoveHead();
	   if(survFlag && pB->Survive(SatTest)) 
	     { if(ExampleOption&16) cout<<endl<< "Sending " << pB->DFnum 
		 << " to sleep.";
	       pB->MarkUnVisited();
	       S2.Insert(pB);
	     }
	   else 
	     { if(ExampleOption&16) cout<<endl<< "Killing " << pB->DFnum;
	       pB->MarkOffStack();
	       pB->FreePartition();
	     }
	 } while(pB != this);
      LastMyName = DFnum;
      while(!S2.isEmpty())
	{ pB = S2.RemoveHead();
	  if(pB->isOnStack())
	    { pB->WakeUp();
	      if(pB->Search()) return true;
	    }
	}
    }
  aqs->releaseTransitionsHead(ThisAQSState);
  return(false);
}



bool Bbar::checkProperty() { return(Search()); }

/*****************************************************************/
/*            Display the Bbar Graph Message Definition	         */
/*****************************************************************/

void Bbar::Display()
{
  int i,j;

  
  cout <<  endl<<  endl;

  cout << "StateID:        " << DFnum           << endl <<
	  "AQS state:      " << ThisAQSState    << endl <<
	  "Automata state: " << *automata_index << endl;
  cout << "Local tracked processes: "; 
  for(Node<TempVar*> *mp = prog->tracked.getHead();mp != NULL;mp = mp->Next)
    { i = mp->data->GetIndexOfProcess();
      for(j=0;j<NumberOfProcesses;j++)
	if(Permutations[j]==i)cout<< j<<" ";
    }
  cout << endl << "Local Partition:    " ;  
  for(i=0;i<NumberOfProcesses;++i)
    cout << (partition[Permutations[i]]>>3);

  cout << endl << "ENABLED bit         " ;  
  for(i=0;i<NumberOfProcesses;++i)
    cout << ((partition[Permutations[i]] &4)>>2);
  cout << endl << "EXECUTE bit         " ;  
  for(i=0;i<NumberOfProcesses;++i)
    cout << ((partition[Permutations[i]] &2)>>1);
  cout << endl << "LEADER bit          " ;  
  for(i=0;i<NumberOfProcesses;++i)
    cout << (partition[Permutations[i]] &1);

  cout << endl << "Permutation:        " ;
  for(i=0;i<NumberOfProcesses;++i)
    cout << Permutations[i];

  cout << endl << "The values: " << endl ;
  (*aqs)[ThisAQSState].PrintValueTable();
  cout << " " <<*automata_index;
}

/*****************************************************************/
/*                    Report() Message Definition		 */
/*    Reports on screen the number of states on the Bbar graph.  */
/*****************************************************************/

void Bbar::Report() 
{ }

/*************************************************************************/
/*               Extraction Operator Overloading Definitions             */
/*************************************************************************/


ostream& operator<<(ostream& Cout, Bbar& arg)
{
  arg.Display();
  return(Cout);
}

/*************************************************************************/
/*             Concurrent Software Systems Laboratory (c)1995            */
/*************************************************************************/
