/*************************************************************************/
/*                           PROPERTY OF THE  				 */
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/
/*************************************************************************/
/*                             Header File                               */
/*************************************************************************/

#ifndef BBAR_H_
#define BBAR_H_
#include         "AQS.h"
extern int       FairnessOption;
/************************************************************************/
/* Bbar class					           	        */
/*									*/
/* This is a class that contains four things: 				*/
/*  1.  A state from the AQS (ThisAQSState)				*/
/*  2.  An index of the equivalent Automata state.			*/ 
/*  3.  The Permutations, an array of Perms and their inverses		*/
/*  4.  A list of transitions to other states.				*/
/*									*/
/* The class is created by calling the constructor with a pointer to	*/
/* starting state, and a pointer to Automata stucture.  The two lists	*/
/* are traversed according the the algorithm presented in Emerson/Sistla*/
/* 1/95, except that a DepthFirst search is used instead of bredth	*/
/* first.  This is to use the recursive call to the constructor for	*/
/* the next state.  It should not make a difference in terms of		*/
/* complexity or requirements for storage, it merely uses the system	*/
/* stack to do what Queue does in AQS algorithm.			*/
/************************************************************************/

class Bbar
{
private:		
	long      ThisAQSState;  
	AutoState*automata_index; 
	int      *Permutations; 
	Bbar     *Link;
        int       status;  
        int       DFnum;
        int       LowLink; 
        int      *partition;       // The actual equivalence rel of  proc's
        int       partition_size;  // its size

        static AQS            *aqs;  
 	static Program        *prog;  
        static int             NumberOfProcesses;
public: static int             Bbarcount;
private:static int             LastMyName;  // Unique Id.
        static clock_t         BeginTime;  // Tracks construction time.
	

	 void  InitializePartition();
	 void  FreePartition();
         void  MarkInFinal(); 
         void  MarkOutFinal();
         bool  isFinal();  
         void  MarkSingle(); 
         void  MarkNotSingle();
         bool  isNotSingle();  
         void  MarkVisited(); 
         void  MarkUnVisited();
         void  MarkOffStack();
         void  MarkOnStack();  // receiver has been placed on the stack
         bool  isVisited();    // Returns true if visited = yes
         bool  isOnStack();    // Returns true if the node is on stack
         void  MarkDeleted();
         bool  isDeleted();
         void  MergePartitions(int*, int*);
         void  MergePartitions(int*, int );
         int*  isSatisfactory();
	 void  GetNewPermutations(int*, AQSTransition*,int,int);  
	 Bbar* CheckRepetition(long, AutoState *, int*);
	 bool  PermutationsMatch(int*);
         bool  Survive(int*);
         void  WakeUp(int*);
         void  WakeUp();


public:
               ~Bbar();
                Bbar(AQS*, Program*);  
	        Bbar(long, AutoState *, int*); // index of AQS, automata, perm
	void    Display();  
        void    Report();
        bool    Search();
        bool    checkProperty();
        friend  ostream&    operator<<(ostream&, Bbar&);

};
        ostream& operator<<(ostream&, Bbar&);
#endif
