/*************************************************************************/
/*                           PROPERTY OF THE                             */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/
#include "HashTable.h" 

#define IFACTOR   500
#define RADIX       7

int                                    isPrime(int);
int                              FindNextPrime(int);
/*************************************************************************/
/*                     HashTable MESSAGE DEFINITIONS                     */
/*************************************************************************/
/*************************************************************************/
/*                     Constructor Message Definition                    */
/*************************************************************************/

  HashTable::HashTable(int SymbolTableSize_) 
  {
    int i;
    SymbolTableSize = SymbolTableSize_;
    Size = SymbolTableSize * IFACTOR;
    Occupied = 0;
 
    if( !isPrime(Size) )
        Size = FindNextPrime(Size);
    cout << Size;
    HTarray = new int [Size]; 
    Free = Size; 

    for(i=0; i < Size; i++) 
        HTarray[i] = -1;    
  }

/*************************************************************************/
/*                     Destructor Message Definition                     */
/*************************************************************************/
 
  HashTable::~HashTable()
  {
     delete [] HTarray;
  }


/*************************************************************************/
/*                       Message getSize() Definition                    */
/*************************************************************************/

  int HashTable::getSize() 
  { 
    return(Size); 
  }

/*************************************************************************/
/*                  Overloading "[]" Operator Definition                 */
/*************************************************************************/

  int& HashTable::operator[](int index)
  {
     assert( index <= Size ); 
     return( HTarray[index] );
  }

/*************************************************************************/
/*                       Messsage Hash() Definition                      */
/*************************************************************************/

int HashTable::hash(long CheckSum)
{ int i = CheckSum % Size; 
  i += (i<0)*Size;
  return( i );  
}  

  int HashTable::hash(int *array)
  {
    int                  HashedIndex;
              
     HashedIndex = MultiplyValues(array, SymbolTableSize);
     
     if(HTarray[HashedIndex] == -1 )    // Hit!
     {
        Free--;
        Occupied++;
     }
     return(HashedIndex);
  }  

/*************************************************************************/
/*                       Message Report() Definition                     */
/*************************************************************************/
  
  void HashTable::Report() 
  {
     cout<<endl<<endl; 
     cout<<"*************** HASH TABLE STATUS ****************"<<endl;
     cout<<" Total Size     ----->    "<<Size<<endl; 
     cout<<" Occupied Slots ----->    "<<Occupied<<endl;
     cout<<" Free Slots     ----->    "<<Free<<endl;
     cout<<"**************************************************"<<endl; 
     cout<<endl; 
  }

/*************************************************************************/
/*                   Utility AddValues() Function                        */
/* max = Size of SymbolTable and of Value Table. Multiplied by FACTOR it */
/* gives the size of the Hash table.                                     */
/*************************************************************************/

  int HashTable::AddValues(int *array, int max) 
  { 
      int              i;
      int            sum;
       
      for(i=0, sum=0; i<max; i++) 
          sum += array[i];
     
      return( sum % Size );      
  }
 
/*************************************************************************/
/*                  Utility MultiplyValues() Function                    */
/*************************************************************************/

  int HashTable::MultiplyValues(int *array, int max) 
  {
      int              i;
      int        product;
      
      for(i=0 , product=1; i<max; i++) 
          product = (product * (array[i]+1)) % Size;

      return(product);     
  }
/*************************************************************************/
/*                    Utility RadixHash() Function                       */
/*************************************************************************/

  int HashTable::RadixHash(int *array, int max)
  {
    int       i;
    int   value;
  
    value = 0; 
    for(i=0; i < max; i++) 
        value = ((value * RADIX) + array[i]) % Size;   

    return(value);  
  }

/***************************************************************************/
/*                       Utility isPrime() Function                        */
/*  Returns one (int) in the argument integer number is Prime,else returns */
/*  a (int) zero.                                                          */ 
/***************************************************************************/

  int isPrime(int number)
  {
    int     div;
    int   prime;

     for(div = 2, prime = 1; div*div <= number; div++)
     {
         if( number % div == 0 )
             prime = 0;     // If divisors detected return false(i.e. 0)
     }
         if(prime)
            return(1);
         else
            return(0);
  }

/***************************************************************************/
/*                     Utility FindNextPrime() Function                    */
/*      Returns the next minimum Prime number than the integer argument.   */
/***************************************************************************/

  int FindNextPrime(int number)
  {
    int   i;

          i = number;
          i++;

          while(!isPrime(i))
                i++;

          return(i);     // Immediately next min prime.
   }

/*************************************************************************/
/*                       Operator "<<" Overloading                       */
/*  The outputted values correspond to indeces on the BaseObject array.  */
/*************************************************************************/
 
  ostream& operator<<(ostream &Cout, HashTable &rhs)
  {
     int          i; 
     int    counter;          

      for(i=0,counter=0; i < rhs.Size; i++) 
      {
           cout<<"["<<i<<"]"<<rhs.HTarray[i]<<" "; 
           counter++;

           if( (counter % 15) == 0 )
           {
               cout<<endl;
               counter = 0;
           }
      }
     rhs.Report();     
     return(Cout);
  }

