/*************************************************************************/
/*                           PROPERTY OF THE                             */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/
#include  "BBar.h"
#include  <fstream.h>
#include  <unistd.h>
#include  <iostream.h>

int             ExampleOption  = 1;
int             SymmetryOption = 3;
int             OnFlyOption    = 1;
int             FairnessOption = 'w';

extern Program  prog;
extern FILE    *yyout, *yyin;

void             memory(char *message = "", int verb = 0);
void             InputInterface(int, char**);
int              yyparse();

clock_t      stat4=0;
clock_t      stat5=0;
clock_t      stat6=0;



int main(int argc, char **argv)
{ clock_t      stat1;
  clock_t      stat2=0;
  clock_t      stat3;
  clock_t      bt, et;       
  int          sum1;
  int          sum2;
  bool         res; 
  

  InputInterface(argc, argv); 
  stat3=clock();

  if(ExampleOption)
    { cout << " *** NumberOfProcesses: " <<prog.GetNumberOfProcesses()<<endl;
      cout << " *** SymmetryOption   : " <<SymmetryOption<<endl;
      cout << " *** OnFlyOption      :" <<(OnFlyOption ? (OnFlyOption > 1 ? 
			" Both AQS nodes and edges onfly":
			" AQS nodes created onfly, edges stored"):
	                " Full AQS created in advance") << endl;
      cout << " *** ExampleOption    :" <<
	(ExampleOption&1?  " GeneralInfo" : "") << 
	(ExampleOption&2?  " Parser" : "") << 
	(ExampleOption&4?  " AQS construction" : "") << 
	(ExampleOption&8?  " Counterexample" : "") <<
	(ExampleOption&16? " BBar construction" : "") <<  endl;
      cout << " *** FairnessOption   :" <<
	(FairnessOption == 's'?" Strong":" Weak")<<  endl;
      if(ExampleOption&2)cout << " *** The Program: " << prog << endl;
      memory("After the Lexical Analysis ----",1);
    }
  bt = clock(); 
  AQS aqs(prog);  // ExampleOption = 4 is used to print info on AQS constr.
  et = clock();
  stat1=(et - bt)/1000000;
  cout<< endl;
  if(ExampleOption) memory("After the AQS Construction ----",1);   

  if(!prog.IsDefined())
    { res = aqs.deadlock();
      cout<<endl<<endl<<endl<<(res?
	" The program CONTAINS a deadlock.":
	" The program DOES NOT CONTAIN a deadlock.") << endl;
      if(ExampleOption)  
	{ cout<<endl<<"           *** REPORT ***"<<endl<<endl;
	  cout<<"Out of that stat4 --------------------> "
	    <<(stat4)<<" sec"<<endl;
	  cout<<"And stat5 ----------------------------> "
	    <<(stat5)<<" sec"<<endl;
	  cout<<"And stat6 ----------------------------> "
	    <<(stat6)/1000000<<" sec"<<endl;
	  cout << endl;
	  sum2 = aqs.Report();
	}
    }
  else
    {
      Bbar  BB(&aqs,&prog);
      bt = clock();
      res = BB.checkProperty();
      et = clock();
      cout<<endl; 
      if(ExampleOption)  memory("At the end of the computation -",1);
      
      cout<<endl<<endl<<endl<<(res?
	" YES - Complement of the Automata property is not satisfied.":
	" NO  - Complement of the Automata property is satisfied.") << endl;
 
      if(ExampleOption)  
	{ cout<<endl<<"           *** REPORT ***"<<endl<<endl;
	  cout<<"AQS Construction Time ----------------------------> "
	    <<stat1<<" sec"<<endl;
	  cout<<"Strongly Connected Components Computation --------> "
	    <<(et - bt)/1000000<<" sec"<<endl;
	  cout<<"Out of that stat4 --------------------> "
	    <<(stat4)<<" sec"<<endl;
	  cout<<"And stat5 ----------------------------> "
	    <<(stat5)<<" sec"<<endl;
	  cout<<"And stat6 ----------------------------> "
	    <<(stat6)/1000000<<" sec"<<endl;
	  cout << endl;
	  sum2 = aqs.Report();
	  
	  cout<<"Number of Bbar states ---------------------> "
	    <<Bbar::Bbarcount<<endl;
	  cout<<"Size of a Bbar state (bytes)---------------> "
	    <<sizeof(Bbar)<<endl;
	  sum1 = (sizeof(Bbar)*Bbar::Bbarcount)/1000;
	  sum2+= sum1;
	  cout<<"Total size of the Bbar states (kbytes) ===========> "
	    <<sum1<<endl;
	 
	}
    }

  cout<<"-------------------------------------------------------------"<<endl; 
  cout<<"Total Computation time ---------------------------> "
    <<(clock()-stat3)/1000000<<" sec"<<endl;
  cout<<"Total size of all structures  (kbytes) ===========> "
    <<sum2<<endl<<endl<<endl;
  cout<<" End of Computation."<<endl;
  cout<<" CSS Laboratory - (c)1996"<<endl<<endl; 
}

void memory(char *message, int verb)
{ static int st=1;
  static long szmax = 0, rssmax = 0;
  static char command[75];
  unsigned long sz, rss;
  static char store[20];

  if(st)
    { st= 0;
      sprintf(command,"rm -f tmp; ps -u | grep %i | grep -v grep >tmp",getpid());
    }
  system(command);
  ifstream file("tmp");
  file.ignore(24)>>sz>>rss;
  file.close();

  if(sz > 999999) 
    { 
      sprintf(store, "%u",sz);
      rss = atoi(store +(strlen(store)/2));
      store[strlen(store)/2]= ' ';
      sz=atoi(store);
    }
  if(sz>szmax) szmax=sz;
  if(rss>rssmax) rssmax=rss;
  if (verb)              
 cout << "MEMORY REPORT: "<< message                 <<"----> "
   << szmax << ", " << rssmax  << endl;
}


void InputInterface(int argc, char **argv) 
{ char  *arg;
//  char  *autom = NULL;
  char  *progr = NULL;
  int   client = 10;
  const char  helptext[] = "Usage: SMC [-vfso] [-p<prog_file>]\n";


  argc--;
  argv += 1;

  for(arg = argv[0];argc;argc--,argv += 1,arg = argv[0])
    { if(arg[0]=='-')switch(arg[1])
        { case 0  : break;
	  case 'v': ExampleOption  = atoi(arg+2); break;
	  case 'o': OnFlyOption    = atoi(arg+2); break;
	  case 's': SymmetryOption = atoi(arg+2); break;
	  case 'f': FairnessOption = arg[2]; break;
	  case 'p': progr = arg+2; break;		    
//	  case 'a': autom = arg+2; break;	
//	  case 'c': client = atoi(arg+2); break;
	  default : cout << helptext; exit(0); 
	}  
      else if(progr == NULL) progr = arg;
//      else if(autom == NULL) autom = arg;
    }
  
  if(progr == NULL) progr =  "ClientServer.prg";
//  if(autom == NULL) autom =  "automata";

  yyout = fopen("tmp", "w");
  yyin = fopen(progr, "r");
  if(yyin == NULL) { cout << "Cannot open program file " << progr<<endl; 
                     exit(0); }
  yyparse();
  fclose(yyin);
//  NumberOfProcesses =  Processes.getSize();
//  yyin = fopen(autom, "r");
//  if(yyin == NULL) {cout << "Cannot open automata file " << autom; exit(0);}
//  yyparse();
//  fclose(yyin);
}





