/***********************************************************************/
/*                           PROPERTY OF THE                           */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY               */
/***********************************************************************/

#include     "Interface.h"
#include  <unistd.h>

//extern clock_t      stat4;
//extern clock_t      stat5;
//extern clock_t      stat6;
//clock_t             st;
//long stat4, stat5,stat6,st;
/***********************************************************************
*
*             ExprNode handels the following cases
*
* data *left    *right   remark      //&|NEGgLl+-/%*,= SPO en M! v V F
* 
* &     expr     expr    log. or
* |     expr     expr    log. and
* N     expr     expr    notequal
* E     expr     expr    equal
* G     expr     expr    greatequal
* g     expr     expr    greater
* L     expr     expr    lessequal
* l     expr     expr    less
* +     expr     expr    sum
* -     expr     expr    binary minus
* /     expr     expr    divison
* %     expr     expr    modulo
* *     expr     expr    product
* ,     expr     expr    action: actions on both side
* =     expr     expr    action: left is of 'V' type
* S     TempVar  expr    big sum
* P     TempVar  expr    big product
* O     TempVar  expr    big or
* e     TempVar  TempVar index equal
* n     TempVar  TempVar index notequal
* M     expr     NULL    unary minus
* !     expr     NULL    log. not
* v     int      NULL    a value
* V     Variable List<TempVar*> a variable and its indeces
* F     Module   NULL    fairness: default fairness condition for the module
***********************************************************************/



//******************** CLASS ExprNode
ExprNode::ExprNode(int d, ExprNode *lc, ExprNode *rc)    // &|NEGgLl+-/%*,= 
{ data        = d;
  LeftChild   = lc;
  RightChild  = rc;

  if(lc->temps->getSize()>rc->temps->getSize())
    { temps = lc->temps;
      lc->temps = rc->temps;
    }
  else temps = rc->temps;
  
  for(Node<TempVar*> *ptr = lc->temps->getHead(); ptr != NULL; ptr = ptr->Next)
    temps->AppendWithoutRepetition(ptr->data);

  delete (lc->temps);
}  

ExprNode::ExprNode(int d, ExprNode *lc)                  // M!
{ data        = d;
  LeftChild   = lc;
  RightChild  = NULL;
  temps       = lc->temps;
}  

ExprNode::ExprNode(int d, TempVar  *lc, ExprNode *rc)    // SPO 
{ data        = d;
  LeftChild   = lc;
  RightChild  = rc;
  temps       = rc->temps;
  temps->Remove(lc);
}  

ExprNode::ExprNode(int d, TempVar  *lc, TempVar  *rc)    // en
{ data        = d;
  LeftChild   = lc;
  RightChild  = rc;
  temps       = new List<TempVar*>;
  temps->Append(lc);
  if(rc->name != NULL)temps->Append(rc);
}  

ExprNode::ExprNode(int val)                              // v 
{ data        = 'v';
  LeftChild   = new int(val);
  RightChild  = NULL;
  temps       = new List<TempVar*>;
}  

ExprNode::ExprNode(Module   *lc)                         // F
{ data        = 'F';
  LeftChild   = lc;
  RightChild  = NULL;
  temps       = new List<TempVar*>;
}  

ExprNode::ExprNode(Variable *lc, List<TempVar*>* rc)     // V 
{ data        = 'V';
  LeftChild   = lc;
  RightChild  = rc;
  temps       = new List<TempVar*>(*rc); 
  for(Node<TempVar*>*t=temps->getHead();t != NULL; t=t->Next)
    if(t->data->name == NULL) temps->Remove(t);
   // It makes a copy that can be freely destroyed later
}  

ExprNode::~ExprNode() { }
List<TempVar*>  *ExprNode::GetTempVars() { return(temps); }

int ExprNode::Evaluate2(int*t, int *p) 
{ // st = clock();
  int i = Evaluate(t, p); 
//  stat5 += (clock() - st); 
  return i;
}
int ExprNode::Evaluate(int*t, int *p) 
{ int      i,j;
  TempVar *tp;

  switch(data) 
    { case '&': return(((ExprNode*)LeftChild) ->Evaluate(t,p) && 
		       ((ExprNode*)RightChild)->Evaluate(t,p));
      case '|': return(((ExprNode*)LeftChild) ->Evaluate(t,p) ||
		       ((ExprNode*)RightChild)->Evaluate(t,p));
      case '+': return(((ExprNode*)LeftChild) ->Evaluate(t,p) + 
		       ((ExprNode*)RightChild)->Evaluate(t,p));
      case '-': return(((ExprNode*)LeftChild) ->Evaluate(t,p) -
		       ((ExprNode*)RightChild)->Evaluate(t,p));
      case '/': return(((ExprNode*)LeftChild) ->Evaluate(t,p) / 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case '%': return(((ExprNode*)LeftChild) ->Evaluate(t,p) % 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case '*': return(((ExprNode*)LeftChild) ->Evaluate(t,p) * 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'N': return(((ExprNode*)LeftChild) ->Evaluate(t,p) != 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'E': return(((ExprNode*)LeftChild) ->Evaluate(t,p) == 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'G': return(((ExprNode*)LeftChild) ->Evaluate(t,p) >= 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'g': return(((ExprNode*)LeftChild) ->Evaluate(t,p) > 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'L': return(((ExprNode*)LeftChild) ->Evaluate(t,p) <= 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'l': return(((ExprNode*)LeftChild) ->Evaluate(t,p) < 
		       ((ExprNode*)RightChild)->Evaluate(t,p)); 
      case 'M': return(-((ExprNode*)LeftChild)->Evaluate(t,p)); 
      case '!': return(!((ExprNode*)LeftChild)->Evaluate(t,p));
      case 'v': return(*(int*)(LeftChild));
      case 'e': return(((TempVar*)LeftChild)->value  == 
		((TempVar*)RightChild)->value);
      case 'n': return(((TempVar*)LeftChild)->value  != 
		((TempVar*)RightChild)->value);
      case 'S': j=0;
                tp = (TempVar*)LeftChild;
                for(tp->value = 0;
		    tp->value < tp->module->num_of_instances;
		    tp->value++)
                 { i  = tp->value;
		   j += ((ExprNode*)RightChild)->Evaluate(t,p);
		   if(tp->value != i)
		     { cout << tp << " occures as bounded and not bounded.";
		       exit(0);
		     }
		 }
                return(j);
      case 'P': j=1;
                tp = (TempVar*)LeftChild;
                for(tp->value = 0;
		    tp->value < tp->module->num_of_instances;
		    tp->value++)
                 { i  = tp->value;
		   j *= ((ExprNode*)RightChild)->Evaluate(t,p);
		   if(tp->value != i)
		     { cout << tp << " occures as bounded and not bounded.";
		       exit(0);
		     }
		 }
                return(j);
      case 'O': j=0;
                tp = (TempVar*)LeftChild;
                for(tp->value = 0;
		    tp->value < tp->module->num_of_instances;
		    tp->value++)
                 { i  = tp->value;
		   j = j || ((ExprNode*)RightChild)->Evaluate(t,p);
		   if(tp->value != i)
		     { cout << tp << " occures as bounded and not bounded.";
		       exit(0);
		     }
		 }
                return(j);
      case 'V': j=0; 
                for(Node<TempVar*> *mp = ((List<TempVar*>*)RightChild)->getHead(); 
		    mp != NULL; mp = mp->Next)
                   j = j * mp->data->module->num_of_instances 
		         + mp->data->GetModifiedValue(p);
		 j += ((Variable*)LeftChild)->offset;
		 return(t[j]);   
	      }
  cout << "Illegal (F,=) node in expression tree"; exit(0);
}

void ExprNode::TakeAction(int*destination,int*departure) 
{ int      j;
  TempVar *tp;

  switch(data) 
    { case 'P': tp = (TempVar*)LeftChild;
                for(tp->value = 0;
		    tp->value < tp->module->num_of_instances;
		    tp->value++)
                 { j  = tp->value;
		   ((ExprNode*)RightChild)->TakeAction(destination,departure);
		   if(tp->value != j)
		     { cout << tp << " occures as bounded and not bounded.";
		       exit(0);
		     }
		 }
                return;
      
      case '=': j=0; 
		Node<TempVar*> *mp;
                for(mp = ((List<TempVar*>*)(((ExprNode*)LeftChild)->RightChild))
		         ->getHead(); mp != NULL; mp = mp->Next)
                   j = j * mp->data->module->num_of_instances 
		         + mp->data->GetModifiedValue(NULL);
		j += ((Variable*)(((ExprNode*)LeftChild)->LeftChild))->offset;
                destination[j] = ((ExprNode*)RightChild)->Evaluate(departure);
                return;
      
      case ',': ((ExprNode*)LeftChild) ->TakeAction(destination,departure);
	        ((ExprNode*)RightChild)->TakeAction(destination,departure);  
                return;
	      }
  cout << "Illegal node in expression tree at action"; exit(0);
}

ostream  &operator<<(ostream&Cout,  ExprNode&d)
{ char p[] = "1";
  p[0] = d.data;

  switch(d.data) 
    { case '&': 
      case '|': 
      case '+': 
      case '-':
      case '/': 
      case '%': 
      case '*': 
      case ',': 
      case '=': Cout <<"("<< *(ExprNode*)(d.LeftChild)<< " "<< p <<" " 
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'N': Cout << "(" << *(ExprNode*)(d.LeftChild) << " != "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'E': Cout << "(" << *(ExprNode*)(d.LeftChild) << " == "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'G': Cout << "(" << *(ExprNode*)(d.LeftChild) << " >= "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'g': Cout <<"(" << *(ExprNode*)(d.LeftChild) << " > "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'L': Cout <<"(" << *(ExprNode*)(d.LeftChild) << " <= "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'l': Cout <<"(" << *(ExprNode*)(d.LeftChild) << " < " 
	        << *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'S': Cout <<"SUM("<< *(TempVar*)(d.LeftChild) << " : "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'P': Cout <<"PRODUCT(" << *(TempVar*)(d.LeftChild) << " : "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'O': Cout <<"OR(" << *(TempVar*)(d.LeftChild) << " : "
		<< *(ExprNode*)(d.RightChild) << ")";
		break;
      case 'e': Cout <<"(" << *(TempVar*)(d.LeftChild) << " == "
		<< *(TempVar*)(d.RightChild) << ")";
		break;
      case 'n': Cout <<"(" << *(TempVar*)(d.LeftChild) << " != "
		<< *(TempVar*)(d.RightChild) << ")";
		break;
      case 'M': Cout <<"( -" << *(ExprNode*)(d.LeftChild) << ")";
		break;
      case '!': Cout <<"( !" << *(ExprNode*)(d.LeftChild) << ")";
		break;
      case 'V': Cout << *(Variable*)(d.LeftChild) << "["
		<< *(List<TempVar*>*)(d.RightChild) << "]";
		break;
      case 'v': Cout << *(int*)(d.LeftChild);
		break;
      case 'F': Cout <<"Default for " << *(Module*)(d.LeftChild);
		break;
	      }
  return(Cout);

}

//******************** CLASS Module
Module::Module(char* name_, int num_of_instances_, int offset_)
{ name = name_; num_of_instances = num_of_instances_; offset = offset_; }

 ostream &operator<<(ostream&Cout,  Module&d)
{ Cout << d.name;  return(Cout);}

//******************** CLASS TempVar
TempVar::TempVar(Module* module_, char* name_) 
{ module = module_; name = name_; value = 0; }

TempVar::TempVar(Module* module_, int value_)
{ module = module_; name = NULL; value = value_; }
               
TempVar::TempVar(TempVar&arg)
{ (*this) = arg; }

TempVar &TempVar::operator = (TempVar&arg)
{ module = arg.module;
  name   = arg.name;
  value  = arg.value; 
  return(arg);
}


ostream  &operator<<(ostream&Cout, TempVar &d)
{ if(d.name != NULL) Cout << d.name;
  Cout<<"("<<d.value<<") of " << *d.module;  
  return(Cout);
}

int TempVar::GetIndexOfProcess()
{ if(module->num_of_instances <= value)
    { cout << *(this) <<" is out of range"; exit(0); } 
  return(module->offset + value); 
}

int TempVar::GetModifiedValue(int *p)
{ int index = GetIndexOfProcess();
  int j;

  if(p == NULL) return(value);
  for(j = module->offset; 
      p[j] != index && j< module->offset + module->num_of_instances; j++);
  if(j == module->offset + module->num_of_instances)
     { cout << *(this) <<" is out of range when permuted"; exit(0); } 
  return(j - module->offset);
}

//******************** CLASS Variable 
Variable::Variable(char* name_,  List<Module *>*indeces_, int value_)
{ offset  = 0;
  value   = value_;
  name    = name_;
  indeces = indeces_;
}

Node<Module *>  *Variable::GetHead()
{ return(indeces->getHead()); }

 ostream  &operator<<(ostream&Cout, Variable &d)
{ Cout << d.name;  return(Cout);}

//******************** CLASS Transition
Transition::Transition(TempVar *module_, ExprNode *action_)
{ module = module_;
  action = action_;
  action->temps->AppendWithoutRepetition(module);
}

ostream  &operator<<(ostream&Cout,  Transition&d)
{ Cout << "Transition of module " << *d.module << endl;
  Cout << *(ExprNode*)(d.action->LeftChild) << " => "
    << *(ExprNode*)(d.action->RightChild)<<endl;
  Cout << "Unbounded indeces: "<< *(d.action->temps)<< endl;
  Cout << "Process that caused the transition is " << d.module->GetIndexOfProcess()<<endl;
 return(Cout);
}

TempVar *Transition::GetProcess()
{ return(module); }
 
bool Transition::TakeAction2(int*destination,int*departure,int size) 
{ // st = clock();
  bool ret = TakeAction(destination,departure,size) ;
//  stat6 += (clock() - st); 
  return ret;
}

extern clock_t      stat4;
bool Transition::TakeAction(int*destination, int* departure,int size)
{ 
  if(((ExprNode*)(action->LeftChild))->Evaluate(departure) == 0) return(false);
  memcpy(destination,departure,size * sizeof(int));
  ((ExprNode*)(action->RightChild))->TakeAction(destination,departure);
  stat4++;
  return(true);
}

//******************** CLASS FairPair
FairPair::FairPair(int type_, ExprNode *pair_)
{ offset  = 0; type = type_; pair = pair_; }

 ostream &operator<<(ostream&Cout,  FairPair&d)
{ Cout << "Fairness type is " << (d.type == 'w'?"weak.":"strong.")<<endl;
  Cout << "Enabled condition: "<<*(ExprNode*)(d.pair->LeftChild)<<endl;
  Cout << "Execute condition: "<<*(ExprNode*)(d.pair->RightChild)<<endl;
 return(Cout); 
}

//******************** CLASS FairCondition
            
void  FairCondition::AddFairPair(int type_,ExprNode* enable, ExprNode* exec)
{ fair_pairs.Insert(new FairPair(type_, new ExprNode(',',enable, exec))); }

 ostream  &operator<<(ostream&Cout,  FairCondition&d)
{ Cout << endl <<"The fairness conditions are the following:"<< endl
    << d.fair_pairs<< endl;
  return(Cout);
}
//******************** CLASS AutoTransition
 ostream &operator<<(ostream&Cout,  AutoTransition&d)
{ Cout << "Transition from " << *d.departure << " to " << *d.destination
    << " on condition " << *(d.condition)<<endl;
  return(Cout); 
}

//******************** CLASS AutoState
AutoState::AutoState(char *name_) { name  = name_; final = false; }
void AutoState::SetFinal()        { final = true;  }
bool AutoState::isFinal()         { return(final); }

 ostream  &operator<<(ostream&Cout,  AutoState&d)
{ Cout << d.name << (d.final?"(F)":"");
  return(Cout);
}

//******************** CLASS Const  

Const::Const(char* name_, int value_)
{ name  = name_;
  value = value_;
}

//******************** CLASS Automata
void            Automata::AddAutoTransition(AutoTransition *t)
{ a_transitions.Append(t);}

AutoState      *Automata::AddAutoState(char *name_)
{ for(Node<AutoState*> *mp = a_states.getHead();mp != NULL;mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) return(mp->data); 
  AutoState *as = new AutoState(strcpy(new char[strlen(name_)+1],name_));
  a_states.Append(as);
  return(as);
}

void            Automata::AddTracked(TempVar *t)
{ for(Node<TempVar*> *mp = tracked.getHead();mp != NULL;mp = mp->Next)
    if(mp->data->module == t->module && mp->data->value == t->value) return;
  tracked.Append(t); 
}

void            Automata::AddTracked(Module *m, int i)
{ for(Node<TempVar*> *mp = tracked.getHead();mp != NULL;mp = mp->Next)
    if(mp->data->module == m && mp->data->value == i) return;
  TempVar *t =new TempVar(m,i);
  tracked.Append(t);
}

Automata::Automata() { }
AutoTransition *Automata::GetTransitions()    { return NULL; }
AutoState      *Automata::GetFirstAutoState() 
                        { return(a_states.getHead()->data); }
bool            Automata::IsDefined()
                        { return(a_states.getSize()?true:false); }
 
ostream &operator<<(ostream&Cout,  Automata&d) 
{ Cout << endl  << "The automata states are: " << d.a_states << endl
    << "The automata has the following transitions: " << endl
    << d.a_transitions <<endl
    << "The tracked processes are: " << d.tracked << endl;
   
  
 return(Cout);
}

//******************** CLASS Program
Program::Program()           
{ table_size       = 0;     
  iterator_stage   = NULL;
  num_of_processes = 0;
}

Module            *Program::FindModule(char *name_)
{ for(Node<Module*> *mp = modules.getHead();
      mp != NULL;
      mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) return(mp->data);
  return(NULL);
}

Const             *Program::FindConst(char *name_)
{ for(Node<Const*> *mp = consts.getHead();
      mp != NULL;
      mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) return(mp->data);
  return(NULL);
}
  
Variable          *Program::FindVariable(char *name_)
{ for(Node<Variable*> *mp = variables.getHead();
      mp != NULL;
      mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) return(mp->data);
  return(NULL);
}

void               Program::AddModule(char* name_, int num_of_instances_)
{ Module *m = new Module(name_,num_of_instances_,num_of_processes);
  modules.Append(m); 
  num_of_processes += num_of_instances_;
}

void               Program::AddConst(char* name_, int value_)
{ consts.Append(new Const(name_,value_)); 
}

void               Program::AddVariable(Variable *v)
{ int i=1;
  Node<Module*> *mp;

  for(mp = v->GetHead(); mp != NULL;mp = mp->Next)
    i*=mp->data->num_of_instances;
  v->size     = i;
  v->offset   = table_size;
  table_size += i;
  variables.Append(v); 
}

void               Program::AddTransition(Transition *t)
{ transitions.Append(t); }

TempVar           *Program::AddTempVar(Module *module_, char *name_)
{ for(Node<TempVar*> *mp = temp_store.getHead();
      mp != NULL;
      mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) 
      { delete [] name_;
	if(mp->data->module == module_) return(mp->data); 
	else return(NULL);
      }
  TempVar *t = new TempVar(module_, name_);
  temp_store.Append(t);
  return(t);
}

TempVar           *Program::AddTempVar(char *name_)
{ for(Node<TempVar*> *mp = temp_store.getHead();
      mp != NULL;
      mp = mp->Next)
    if(strcmp(mp->data->name, name_)==0) 
      { delete [] name_;
	return(mp->data);
      } 
  delete [] name_;
  return(NULL);
}

int Program::GetTableSize()                  { return(table_size); }
int Program::GetNumberOfProcesses()          { return(num_of_processes); }
Node<Variable*> *Program::GetVariablesHead() { return(variables.getHead()); }

int Program::GetModulePartition(int* part)       
{ static int  flag = 1;
  static int *ModulePartition;
  static int  i = 0;
  if(flag)
    { flag = 0;
      ModulePartition = new int[num_of_processes];

      int            n;
      Node<Module*> *mp;
      
      for(mp = modules.getHead(); mp != NULL; mp = mp->Next,i++)
	for(n = mp->data->offset; n < mp->data->offset + 
	    mp->data->num_of_instances; n++)
	  ModulePartition[n] = i;
    }
  memcpy(part, ModulePartition, sizeof(int) * num_of_processes);
  return(i);
}

char** Program::GetValueTableNames()
{ char **names = new char*[table_size];
  char **ret = names;

  for(Node<Variable*>*mp1 = variables.getHead(); mp1 != NULL; mp1 = mp1->Next)
    for(int i = 0; i<mp1->data->size; i++)
      *(names++) = mp1->data->name;

 return(ret);
}


void Program::GetTable(int*table)
{ Node<Variable*> *mp1;
  int              i;

  for(mp1 = variables.getHead(); mp1 != NULL; mp1 = mp1->Next)
    for(i = 0; i<mp1->data->size; i++)
      *(table++) = mp1->data->value;
}

Transition* Program::TransitionIteratorReset()
{ Node<TempVar*> *mp2;
  iterator_stage = transitions.getHead();

  for(mp2 = iterator_stage->data->action->GetTempVars()->getHead(); 
      mp2 != NULL; mp2 = mp2->Next)
    mp2->data->value = 0;
  return(iterator_stage->data);
}
  

Transition* Program::TransitionIteratorNext()
{ Node<TempVar*> *mp2;
  
  if(iterator_stage == NULL) return(NULL);
  for(mp2 = iterator_stage->data->action->GetTempVars()->getHead(); 
      mp2 != NULL; mp2 = mp2->Next)
    { if(mp2->data->value == mp2->data->module->num_of_instances)
	{ cout << *(mp2->data) << " was used as bounded and unbounded " 
	       << "in the same expression"; exit(0); }
     if(++mp2->data->value < mp2->data->module->num_of_instances)
       return(iterator_stage->data); 
      mp2->data->value = 0;
    }

  iterator_stage = iterator_stage->Next;
  if(iterator_stage == NULL) return(NULL);

  for(mp2 = iterator_stage->data->action->GetTempVars()->getHead(); 
      mp2 != NULL; mp2 = mp2->Next)
    mp2->data->value = 0;  
  return(iterator_stage->data);
}

ostream     &operator<<(ostream&Cout,  Program&d)
{ Cout << "The program consists of the following transitions:"<<endl
    << d.transitions << *(FairCondition*)&d << *(Automata*)&d;
  return(Cout);
}


