/***********************************************************************/
/*                           PROPERTY OF THE                           */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY               */
/***********************************************************************/

#ifndef ENUM_BOOLEAN_
#define ENUM_BOOLEAN_
   enum bool { false, true };
#endif 

#include     <assert.h>
#include     <iostream.h>
#include     <stdio.h>
#include     <stddef.h>
#include     <stdlib.h>
#include     <string.h>

#include     "List.h"

class TempVar;
class Module;
class Variable;

/***********************************************************************
*
*             ExprNode handels the following cases
*
* data *left    *right   remark      //&|NEGgLl+-/%*,= SPO en M! v V F
* 
* &     expr     expr    log. or
* |     expr     expr    log. and
* N     expr     expr    notequal
* E     expr     expr    equal
* G     expr     expr    greatequal
* g     expr     expr    greater
* L     expr     expr    lessequal
* l     expr     expr    less
* +     expr     expr    sum
* -     expr     expr    binary minus
* /     expr     expr    divison
* %     expr     expr    modulo
* *     expr     expr    product
* ,     expr     expr    action: actions on both side
* =     expr     expr    action: left is of 'V' type
* S     TempVar  expr    big sum
* P     TempVar  expr    big product
* O     TempVar  expr    big or
* e     TempVar  TempVar index equal
* n     TempVar  TempVar index notequal
* M     expr     NULL    unary minus
* !     expr     NULL    log. not
* v     int      NULL    a value
* V     Variable List<TempVar*> a variable and its indeces
* F     Module   NULL    fairness: default fairness condition for the module
***********************************************************************/


class   ExprNode
{ private:
  int              data;
  void            *LeftChild;
  void            *RightChild;
  List<TempVar*>  *temps;

   

  public:
   ExprNode(int, ExprNode*, ExprNode*);
   ExprNode(int, ExprNode*);
   ExprNode(int, TempVar*, ExprNode*);
   ExprNode(int, TempVar*, TempVar*);
   ExprNode(int);
   ExprNode(Module*);
   ExprNode(Variable*,List<TempVar*>*);

  ~ExprNode();
  int              Evaluate(int*,int *p=NULL);
  int              Evaluate2(int*,int *p=NULL);
  void             TakeAction(int*,int*);
  List<TempVar*>  *GetTempVars();
  friend ostream  &operator<<(ostream&,  ExprNode&);
  friend ostream  &operator<<(ostream&,  class Transition&);
  friend ostream  &operator<<(ostream&,  class FairPair&);
  friend ostream  &operator<<(ostream&,  class AutoTransition&);
  friend class Transition;
};


class Module
{ public:
  char            *name;   
  int              num_of_instances;
  int              offset;

                   Module(char*, int, int);
  friend ostream &operator<<(ostream&,  Module&);
};



class TempVar
{ public:
  
  Module          *module;
  char            *name;
  int              value;
 
                   TempVar(Module*, char*);  
                   TempVar(Module*, int);
                   TempVar(TempVar&);         
  TempVar         &operator = (TempVar&);
  int              GetIndexOfProcess();
  int              GetModifiedValue(int*);
  friend ostream  &operator<<(ostream&, TempVar&);
};


class Variable 
{ public:
  int              offset;
  int              size;
  int              value;
  char            *name;
  List<Module *>  *indeces;

                   Variable(char*,  List<Module *>*, int);
  Node<Module *>  *GetHead();
  friend ostream  &operator<<(ostream&, Variable &);
};


class Transition
{ public:

  TempVar         *module;
  ExprNode        *action;
 
                   Transition(TempVar*, ExprNode*);
  TempVar         *GetProcess();
  bool             TakeAction(int*,int*,int);
  bool             TakeAction2(int*,int*,int);

  friend ostream  &operator<<(ostream&,  Transition&);
};


class FairPair
{ public: 
 
  int              type;
  int              offset;
  ExprNode*        pair;
                   FairPair(int, ExprNode*);
  friend ostream &operator<<(ostream&,  FairPair&);
};

class FairCondition
{ public:

  List<FairPair*>  fair_pairs;             
  void             AddFairPair(int, ExprNode*, ExprNode*);
  friend ostream  &operator<<(ostream&,  FairCondition&);
};


class AutoState
{ private:
  char            *name;
  bool             final;
  public:
                   AutoState(char*); 
  void             SetFinal();
  bool             isFinal();
  friend class     Automata;
  friend ostream  &operator<<(ostream&,  AutoState&);
};


class AutoTransition
{ public:
  ExprNode        *condition;
  AutoState       *departure;
  AutoState       *destination;
  
  friend ostream &operator<<(ostream&,  AutoTransition&);
};

class Const
{  public:  Const(char*, int);
     char  *name;
     int    value;
};

class Automata
{  private:
     List<AutoTransition*>  a_transitions;
     List<AutoState*>       a_states;
     List<TempVar  *>       tracked;

   public:
     void            AddAutoTransition(AutoTransition*);
     AutoState      *AddAutoState(char*);
     void            AddTracked(Module*,int);
     void            AddTracked(TempVar*);

                     Automata();
     AutoTransition *GetTransitions();
     AutoState      *GetFirstAutoState();
     bool            IsDefined();
     friend ostream &operator<<(ostream&,  Automata&);
     friend class Bbar;
};



class Program : public FairCondition, public Automata
{ private:

  List<Module    *>  modules;
  List<Const     *>  consts;
  List<Variable  *>  variables;
  List<Transition*>  transitions;           
  List<TempVar   *>  temp_store;            
  Node<Transition*> *iterator_stage;
  int                num_of_processes;
  int                table_size;

public:             Program();
  Const             *FindConst(char*);
  Module            *FindModule(char*);
  Variable          *FindVariable(char*);
  void               AddConst(char*,int);
  void               AddModule(char*, int);
  void               AddVariable(Variable*);
  void               AddTransition(Transition*);
  TempVar           *AddTempVar(Module*, char*);
  TempVar           *AddTempVar(char*);
  int                GetTableSize();
  int                GetNumberOfProcesses();
  char             **GetValueTableNames();
  void               GetTable(int*);
  Node<Variable*>   *GetVariablesHead();                  
  int                GetModulePartition(int*);            
  Transition        *TransitionIteratorReset();
  Transition        *TransitionIteratorNext();
  friend ostream     &operator<<(ostream&,  Program&);
};
