/***********************************************************************/
/*                           PROPERTY OF THE                           */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY               */
/***********************************************************************/
#include "List.h"

/***********************************************************************/
/*                NODE OBJECT - TEMPLATE CLASS DEFINITION              */
/***********************************************************************/
/***********************************************************************/
/*                       OBJECT'S CONSTRUCTOR                          */
/***********************************************************************/
  template<class Type>
  Node<Type>::Node(Type &info) 
  {
    data = info; 
    Next = NULL;  
  }

/***********************************************************************/
/*                    getData() MESSAGE DEFINITION                     */
/***********************************************************************/

  template<class Type>
  Type Node<Type>::getData()
  {
    return(data);
  }

/***********************************************************************/
/*           LINKED LIST OBJECT - TEMPLATE CLASS DEFINITION            */
/***********************************************************************/
/***********************************************************************/
/*                   DEFAULT CONSTRUCTOR DEFINITION                    */
/***********************************************************************/

   template<class Type> 
   List<Type>::List() 
   {
     Head = NULL;
     Tail = NULL;
     Size =    0;
   }
/***********************************************************************/
/*               CONSTRUCTOR FOR TYPE OBJECTS DEFINITION               */ 
/***********************************************************************/

  template<class Type>
  List<Type>::List(Type &object)
  {
    Append(object);
  }

/***********************************************************************/
/*              CONSTRUCTOR FOR LIST OBJECTS DEFINITION                */
/*  This message sets the limits of the range of the receiver Variable */
/*  object which is an Indexed variable.                               */
/* In this message,the argument for the constructor is a List object.  */
/* This message is coupled with the '=' operator.                      */
/***********************************************************************/

  template<class Type>
  List<Type>::List(List<Type> &object)
  {
     Head = NULL;
     Tail = NULL;
     Size =    0;
    (*this) = object;
  }

/***********************************************************************/
/*                    DEFAULT DESTRUCTOR DEFINITION                    */
/***********************************************************************/
  
  template<class Type>
  List<Type>::~List()
  {
    if(!isEmpty())       // Only on Non-empty lists.
       Deallocate();
  }
 
/***********************************************************************/
/*                    isEmpty() MESSAGE DEFINITION                     */
/*                 Returns TRUE if empty, FALSE if not.                */
/***********************************************************************/

  template<class Type> 
  bool List<Type>::isEmpty()
  {
    if(Head == NULL) 
       return(true);
    else
       return(false); 
  } 

/***********************************************************************/
/*                    Display() MESSAGE DEFINITION                     */
/* For primitive types will work. For user defined Type the user has   */
/* to define the order that the output goes.                           */   
/***********************************************************************/
 
  template<class Type> 
  void List<Type>::Display() 
  { 
    Node<Type> *NodePtr = Head; 

      if(isEmpty())           // Empty list case.
      {
        
        return;
      }
 
      if( Size == 1 )         // List of size equal to 1 case.
      {
        cout<<NodePtr->data<<flush;
        return;
      } 
 
                             
      while(NodePtr != NULL)  // All other cases.
      {
        cout<<NodePtr->data<<", "<<flush;
        NodePtr=NodePtr->Next;
      }      
  }


/***********************************************************************/
/*               CreateNode() UTILITY MESSAGE DEFINITION               */
/*      Creates a node of Type data type and returns its address.      */
/***********************************************************************/ 
 
  template<class Type> 
  Node<Type>* List<Type>::CreateNode(Type &object)  
  {
    Node<Type>  *NewPtr; 
    NewPtr = new Node<Type>(object);   // Invoke the Node Constructor.
    assert(NewPtr != 0);               // Should never return NULL.
    return(NewPtr);                    // Return the nodes' address. 
  }


/***********************************************************************/
/*                    Append() MESSAGE DEFINITION                      */
/*              Appends a new node at the end of the list.             */
/***********************************************************************/    
 
template<class Type> 
List<Type>& List<Type>::Append(Type &object)
{
  Node<Type>          *NewPtr;
  
  NewPtr = CreateNode(object); 
  
  if(!Size)  // If the list is empty, then link the first node.
    {
      Head = NewPtr;
      Tail = NewPtr;
    }
  
  else
    {
      Tail->Next =     NewPtr;
      Tail       = Tail->Next;
    }
  
  Size++;        // Increase the Size of the list. 
  return(*this); // For cascaded Appends.   
}

template<class Type> 
List<Type>& List<Type>::AppendWithoutRepetition(Type &object)
{ Node<Type> *ptr;
  
  for(ptr = Head; ptr != NULL; ptr = ptr->Next)
    if( ptr->data == object) return(*this);
  return(Append(object));
}

/**********************************************************************/
/*                    Insert() MESSAGE DEFINITION                     */
/*          Inserts a Node from the FRONT side of the list.           */
/**********************************************************************/

template<class Type>
List<Type>& List<Type>::Insert(Type &object) 
{
  Node<Type>          *NewPtr; 
  
  NewPtr = CreateNode(object);         
  
  if(isEmpty())    // Empty list case.
    {
      Head = NewPtr;  // Link the first Node.
      Tail = NewPtr;  // Tail points to new node.
    }
  
  else
    {
      NewPtr->Next =   Head; // New Node is linked.
      Head         = NewPtr; // Head points to the new Node.
    }
  Size++;               // Increase the Size of the list.    
  return(*this);        // For cascaded Inserts.
}       
    
   
/**********************************************************************/
/*                    getSize() MESSAGE DEFINITION                    */
/*   Returns the total number of nodes from Head to Tail locations    */
/**********************************************************************/
    
  template<class Type> 
  int List<Type>::getSize()
  {
    return(Size); 
  }
   
/**********************************************************************/
/*                    getTail() MESSAGE DEFINITION                    */
/*   Returns the address of the last Node added on the Linked List.   */
/**********************************************************************/  
    
   template<class Type>
   Node<Type>* List<Type>::getTail()
   { 
     return(Tail);
   }

/**********************************************************************/
/*                    getHead() MESSAGE DEFINITION                    */
/*   Returns the address of the last Node added on the Linked List.   */
/**********************************************************************/  
    
   template<class Type>
   Node<Type>* List<Type>::getHead()
   { 
     return(Head);
   }

    
/***********************************************************************/
/*                    UTILITY Deallocate() FUNCTION                    */
/*  Behaves as a destructor for a list object, and is called manually. */
/***********************************************************************/
  
  template <class Type>
  void List<Type>::Deallocate()
  {
    Node<Type>  *p;   // Utility Pointer serving as Predecessor to the 
                      // Successor pointer, which is the Head pointer.    
    p = Head;

    if( !isEmpty() )  // If non-empty Receiver List.
    {
		 while(Head != NULL)
		{
		  Head = Head->Next;
		  delete p;
		  p = Head;
		}
	 }

	Head = Tail = NULL;
	Size =    0;  // Size is cleared as well.
  }


/***********************************************************************/
/*                  UTILITY FindPredecessor() FUNCTION                 */
/* Returns a (non-NULL)pointer to the predecessor of the argument node.*/
/* The function either returns a non NULL pointer to the predecessor   */
/* node, or returns NULL, in case the argument pointer does not exist  */
/* on the linked list.					               */
/***********************************************************************/

template<class Type>
Node<Type>*  List<Type>::FindPredecessor(Node<Type> *pN)
{
  Node<Type> *ptr;
  
  for(ptr = Head; ptr != NULL; ptr = ptr->Next)
    if( ptr->Next == pN) return(ptr);
  return(NULL);     // The argument node was not found.
}

/***********************************************************************/
/*                     UTILITY Remove() FUNCTION                       */
/*  Removes the argument node from the list. If the operation succeeds */
/*  then a Boolean True is returned, else a Boolean False is returned. */
/***********************************************************************/

template<class Type>
bool List<Type>::Remove(Node<Type> *pN)
{
  Node<Type>  *pTemp;
  
  if( pN == NULL ) return(false);
  
  if( pN == Head ) {  Head = pN->Next;
		      delete pN;
		      Size--;
		      if(!Size) Tail = NULL;
		      return(true);
		    }

  if( Size == 1) return(false);
       
  pTemp = FindPredecessor(pN);

  if(pTemp == NULL) return(false);  // pN does not exist.  
  pTemp->Next = pN->Next;

  if( pN == Tail) Tail = pTemp;
		   
  delete pN;
  Size--;
  return(true);
}

template<class Type> 
bool List<Type>::Remove(Type element)
{
  Node<Type> *ptr; 
  
  // Look for the element first. 
  
  for(ptr = Head; ptr!= NULL; ptr = ptr->Next)
    if( ptr->data == element) break; 
   
  // Call the Remove() message using the pointer.
  return( this->Remove(ptr) ); 
}

template<class Type>
Type List<Type>::RemoveHead()
{ assert(Size);
  Type ret;
  Node<Type> *pN;

  if(Size == 1) Tail = NULL;
  Size--;
  ret  = Head->data;
  pN   = Head;
  Head = pN->Next;
  delete pN;
  return(ret);
}
  
 

/***********************************************************************/
/*                    OVERLOADING OF THE '=' OPERATOR                  */
/*           The default copy constructor will not work here.          */
/***********************************************************************/

  template<class Type>
  List<Type>&  List<Type>::operator=(List<Type> &rhs)
  {
     if( this != &rhs )
     {
	 Node<Type> *rhsPtr = rhs.Head;
	 Deallocate();  // The Receiver's List.

	 while( rhsPtr != NULL)
	 {
		Append(rhsPtr->data);
		rhsPtr = rhsPtr->Next;
	 }
     }
     return(*this);
  }

/***********************************************************************/
/*                   OVERLOADING OF THE "<<" OPERATOR                  */
/* This message is needed so that the list will compile when List      */
/* objects are instantiated where the template argument is a List<Type>*/
/***********************************************************************/

template<class Type>
ostream& operator<<(ostream &Cout, List<Type> &rhs)
{
  
  if( rhs.isEmpty() )      // Empty list case.
    return(Cout);

  Cout<<*(rhs.Head->data)<<flush;
  for(Node<Type> *NodePtr = rhs.Head->Next; 
      NodePtr != NULL;
      NodePtr=NodePtr->Next)  
    Cout<<", "<<*(NodePtr->data)<<flush;
  
  return(Cout);
}

/************************************************************************/
/*          Concurrent Software Systems Laboratory (c)1995              */
/************************************************************************/
