/*************************************************************************/
/*                           PROPERTY OF THE                             */ 
/*                CONCURRENT SOFTWARE SYSTEMS LABORATORY                 */
/*************************************************************************/

#define STATIC 
#include "AQS.h"  
#include "time.h"
extern int          SymmetryOption;
extern clock_t      stat4;
extern clock_t      stat5;
extern clock_t      stat6;
clock_t             st;
 
  

template <class Type> void Sort(Type **t, int n)
{ int i,j;
  Type *tt;
  if(n == 1) return;
  for(i = 1; i<n; i++)
    { j = i;
      while(j && *(t[j]) < *(t[j-1]))
	{ tt     = t[j];
	  t[j]   = t[j-1];
	  t[--j] = tt;
	}
    }
}


class Value
{ private:
    
    int            *partition;
    int            *proc_indexes;
    int             num_of_pi;
    Variable       *var;
    int            *the_value;
    

  public:
    Value(Variable *v, List<TempVar*> *l, int *val)
      { var                 = v; 
	num_of_pi           = l->getSize();
	proc_indexes        = new int[num_of_pi];
	int the_value_index = 0;
	Node<TempVar*> *mp  = l->getHead();
	
	for(int i = 0; i<num_of_pi; i++)
	  { the_value_index  = the_value_index * 
	                       mp->data->module->num_of_instances 
			       + mp->data->value;
	    proc_indexes[i]  = mp->data->GetIndexOfProcess();
	    mp = mp->Next;
	  }
	the_value_index += var->offset;
	the_value = val + the_value_index;
      }
    void Set(int *p){ partition = p; }   
    bool operator != (Value &arg)
      { if(var != arg.var || *the_value != *arg.the_value) return(true);
	for(int i = 0; i<num_of_pi; i++)
	  if(partition[proc_indexes[i]] != 
	     arg.partition[arg.proc_indexes[i]]) return(true);
	return(false);
      }
    bool operator <  (Value &arg)
      { if(var != arg.var) return(false);
	if(*the_value < *arg.the_value) return(true);
	if(*the_value > *arg.the_value) return(false);
	for(int i = 0; i<num_of_pi; i++)
	  { if(partition[proc_indexes[i]] < 
	       arg.partition[arg.proc_indexes[i]]) return(true);
	    if(partition[proc_indexes[i]] > 
	       arg.partition[arg.proc_indexes[i]]) return(false);
	  }
	return(false);
      }
};

class Domain
{ private:
    Value   **values;
    int       size;
    int      *value_table;
    int      *partition;
    int       id;
   
  public:
    Domain(int i, int *v) 
      { id = i; 
	value_table = v;
	size = 0; 
	values = NULL;
      }
    void AddVariable(Variable *v, Node<Module*> *mp2, int myindex)
      { List<TempVar*> *list; 
        Node<TempVar*> *list2;
	Node<Module *> *mp;
	TempVar        *tv;
	int             carry;
	int             reduced_size;


	reduced_size = v->size/mp2->data->num_of_instances;
	Value **tmp  = values;
	values       = new Value*[size + reduced_size];
	if(size) { memcpy(values, tmp, size * sizeof(Value*));
		   delete [] tmp;
		 }
	
	list = new List<TempVar*>;
	for(mp = v->GetHead(); mp != NULL; mp = mp->Next)
	  { tv = new TempVar(mp->data, "dummy");
            if(mp == mp2) { tv->value = myindex;
			    tv->name  = "active";
			  }
	    list->Append(tv);
	  }
	if(reduced_size == 1 ) 
	  { values[size++] = new Value(v,list,value_table); 
	    return;
	  } 
	do { values[size++] = new Value(v,list,value_table);
	     list2 = list->getHead();
	     list  = new List<TempVar*>; 
	     carry = 1;
	     for(;list2 != NULL; list2 = list2->Next)
	       { tv = new TempVar(*(list2->data));      // Copy constructor
		 if(carry && tv->name[0] == 'd')
		   { if(++tv->value == tv->module->num_of_instances)
		       tv->value = 0;
		   else carry = 0;
		   }
		 list->Append(tv);
	       }
	   } while(!carry);
	delete list;
      }
    void Set(int *p)
      { partition = p; 
	for(int i=0; i<size; i++) 
	  values[i]->Set(p);
      }
	 
    void Order()      { Sort(values,size); }
    int  GetId()      { return(id); }
    bool operator != (Domain &arg)
      { if(partition[id] != arg.partition[arg.id]) return(true);
	for(int i=0; i<size; i++)
	  if(*(values[i]) != *(arg.values[i])) return(true);
	return(false);
      }
    bool operator <  (Domain &arg)
      { if(partition[id] < arg.partition[arg.id]) return(true);
	if(partition[id] > arg.partition[arg.id]) return(false);
	for(int i=0; i<size; i++)
	  { if(*(values[i]) < *(arg.values[i])) return(true);
	    if(*(arg.values[i]) < *(values[i])) return(false);
	  }
	return(false);
      }
};

class Domains
{ private:
    Domain **doms;
    int     *value_table;
    int     *partition;
    int      partition_size;       // Classes are numbered 0...p_s-1
    int     *next_partition;
    int      next_partition_size;
    int      NP; 
    int      TS;

  public:
    Domains(Program &prog)
      { int i;
	Node<Variable*> *mp1;
	Node<Module  *> *mp2;
	NP = prog.GetNumberOfProcesses();
        TS = prog.GetTableSize();
	value_table = new int [TS];
	next_partition = new int [NP];
	doms = new Domain*[NP];
	for(i=0; i<NP; i++) doms[i] = new Domain(i,value_table); 

	for(mp1 = prog.GetVariablesHead(); mp1 != NULL; mp1 = mp1->Next)
	  for(mp2 = mp1->data->GetHead(); mp2 != NULL; mp2 = mp2->Next)
	    for(i = 0; i < mp2->data->num_of_instances; i++)
	      doms[mp2->data->offset + i]->AddVariable(mp1->data, mp2, i);

      } 
    void Set(int *v, int *p) 
      { memcpy(value_table,v,sizeof(int) * TS); 
	partition      = p; 
	partition_size = 0;
	for(int i=0; i<NP; i++) 
	  { if(partition_size < p[i]) partition_size = p[i];
	    doms[i]->Set(p);
	  }
	partition_size++;
	memcpy(next_partition, partition, sizeof(int) * NP);
	next_partition_size = partition_size;
      }
    void Sort() { ::Sort(doms,NP); }
    void AdjustPartition()
      { int actual = doms[0]->GetId();
	next_partition[actual] = (actual << 2) + 3;
 
	for(int i = 1; i<NP; i++)
	  { if(*(doms[i-1]) != *(doms[i])) 
	      { actual = doms[i]->GetId();
		next_partition[actual] = (actual << 2) + 3;
	      }
	    else next_partition[doms[i]->GetId()] =  actual << 2;
	  }
	memcpy(partition, next_partition, sizeof(int) * NP);
      }

    void Order()
      { int i;
    	memcpy(partition, next_partition, sizeof(int) * NP);
	partition_size = next_partition_size; 
	for(i=0; i<NP; i++) doms[i]->Order();
	
	::Sort(doms,NP);
	next_partition[doms[0]->GetId()] = 0;
        next_partition_size = 0;
	for(i=1; i<NP; i++) 
	  { if(*(doms[i-1]) != *(doms[i])) next_partition_size++;
	    next_partition[doms[i]->GetId()] = next_partition_size;
	  }
	next_partition_size++;
      }
    bool WasChange() 
      { return((bool)(partition_size != next_partition_size)); }
    int PartitionSize() 
      { return(partition_size); } 
    bool operator != (Domains &arg)
      { for(int i=0; i<NP; i++)
	  if(*(doms[i]) != *(arg.doms[i])) return(true);
	return(false);
      }
    int operator [] (int i)   
      { assert(i<NP); return(doms[i]->GetId()); }
};

bool ComputeFixpoint(int *backup,
		     int *actual,
		     int *b_part,
		     int *a_part,
		     Program &prog,
                     int &part_size)
{ static Domains b_doms(prog);
  static Domains a_doms(prog);
  b_doms.Set(backup,b_part);
  a_doms.Set(actual,a_part);
 	      
  do { b_doms.Order();              // Do the iteration step
       a_doms.Order();              // Do the iteration step
       if(b_doms != a_doms) return(false);
			   
       } while(b_doms.WasChange());
  part_size = a_doms.PartitionSize();
  return(true);
}


void AddRandomPair(int *b_part, int *a_part, int NP)
{ int *tmp = new int[NP];
  int i,j, t, sa = 0, sb = 0;

  for(i = 0; i < NP; i++) tmp[i] = 0; 
  for(i = 0; i < NP; i++) 
    if(++tmp[b_part[i]] == 2)
      { t = b_part[i];
	for(j = 0; j < NP; j++) 
	  { if(b_part[j] > t) b_part[j]++;
	    if(b_part[j] == t)
	      { if(sb) b_part[j]++;
		else sb++;
	      }
	    if(a_part[j] > t) a_part[j]++;
	    if(a_part[j] == t)
	      { if(sa) a_part[j]++;
		else sa++;
	      }
	  }
	break;
      }
  delete [] tmp;
}

bool PermutationFixes(int *SValues1, int *SValues2, int *perm, Program& prog)
{ static int                   flag = 1;            // static is important
  STATIC Node<Variable*>      *nVar;
  STATIC Node<Module *>       *nMod;
  static List<List<TempVar*>*> list_of_lists;       // static is important 
  STATIC Node<List<TempVar*>*>*nLL; 
  STATIC List<TempVar*>       *actual_list;
  STATIC Node<TempVar*>       *nTV;
  STATIC TempVar              *tv;
  static int                  *index_store;         // static is important 
  STATIC int                  *next_index;
  STATIC int                   ind;
  STATIC int                   carry;
  STATIC bool                  ret;
 
  if(flag)
    { index_store = new int[prog.GetTableSize()];
      next_index  = index_store;
      for(nVar = prog.GetVariablesHead(); nVar != NULL; nVar = nVar->Next)
	{ actual_list = new List<TempVar*>;
	  for(nMod = nVar->data->GetHead(); nMod != NULL; nMod = nMod->Next)
	    actual_list->Append(new TempVar(nMod->data, "dummy"));
	  list_of_lists.Append(actual_list);
	  do { carry  = 1;
	       ind    = 0;
	       for(nTV = actual_list->getHead(); nTV != NULL; nTV = nTV->Next)
		 { tv  = nTV->data;
		   ind = ind * tv->module->num_of_instances + tv->value; 
		   if(carry)
		     { if(++tv->value == tv->module->num_of_instances)
			 tv->value = 0;
		       else carry = 0;
		     }
		 }
	       ind += nVar->data->offset;
	       *(next_index++) = ind;      
	     } while(!carry);
	}
      flag = 0;
    }

  ret = true;
  next_index = index_store;
  for(nVar = prog.GetVariablesHead(), nLL = list_of_lists.getHead(); 
      nVar != NULL && ret; 
      nVar = nVar->Next, nLL = nLL->Next)
    do { carry  = 1;
	 ind    = 0;
	 actual_list = nLL->data;
	 for(nTV = actual_list->getHead(); nTV != NULL; nTV = nTV->Next)
	   { tv = nTV->data;
	     ind = ind * tv->module->num_of_instances 
	               + tv->GetModifiedValue(perm);
	     if(carry)
	       { if(++tv->value == tv->module->num_of_instances)
		   tv->value = 0;
	         else carry = 0;
	       }
	   }
	 ind += nVar->data->offset;
	 if(SValues1[ind] != SValues2[*(next_index++)]) 
	   { ret = false;
	     carry = 1;
	     for(nTV = actual_list->getHead(); nTV != NULL; nTV = nTV->Next)
	       nTV->data->value = 0;
	   }
       } while(!carry);
  return ret;
}

bool StateEquivalence0(int*backup,int*actual,int*perm, Program&prog)
{ if(memcmp(backup, actual, sizeof(int)*prog.GetTableSize())== 0)
    return(true);
  return(false);
}


bool StateEquivalence1(int*backup,int*actual,int*perm, Program&prog)
{ static int  flag = 1;
  static int *b_part;
  static int *a_part;
  static int  NP; 
  static Domains b_doms(prog);
  static Domains a_doms(prog);

  if(flag) { NP = prog.GetNumberOfProcesses();
             a_part = new int [NP];
	     b_part = new int [NP];
	     flag = 0;
	   }

  prog.GetModulePartition(b_part);
  prog.GetModulePartition(a_part);

  b_doms.Set(backup,b_part);
  a_doms.Set(actual,a_part);
     
  b_doms.Sort();            
  a_doms.Sort();        
   

  for(int i=0; i<NP; i++) 
      perm[b_doms[i]] = a_doms[i];

  return(PermutationFixes(backup, actual, perm, prog));
 
} 


bool StateEquivalence2(int*backup,int*actual,int*perm, Program&prog)
{ static int  flag = 1;
  static int *b_part;
  static int *a_part;
  static int  NP; 
  static int  NM; 
  static int  count = 0;
  static int  part_size;

  count++;
  if(flag) { NP = prog.GetNumberOfProcesses();
             a_part = new int [NP];
	     b_part = new int [NP];
	     flag = 0;
	   }
  NM = prog.GetModulePartition(a_part);
  NM = prog.GetModulePartition(b_part);

  if(!ComputeFixpoint(backup,actual,b_part,a_part,prog,part_size))
    return(false);

  for(;part_size < NP;part_size++)
    AddRandomPair(b_part, a_part, NP);
 
  for(int i=0; i<NP; i++) for(int j= 0; j<NP; j++)
    if(b_part[i] == a_part[j])
      { perm[i] = j; break; }

  if(PermutationFixes(backup, actual, perm, prog))
    return(true);

  return(false);
} 

bool StateEquivalence3(int*backup,int*actual,int*perm, Program&prog)
{ static int  flag = 1;
  static int *b_part;
  static int *a_part;
  static int  NP; 
  static int  NM; 
  static int  part_size;

  if(flag) { NP = prog.GetNumberOfProcesses();
             a_part = new int [NP];
	     b_part = new int [NP];
	     flag = 0;
	   }
  NM = prog.GetModulePartition(a_part);
  NM = prog.GetModulePartition(b_part);
 
  for(part_size = NM;
      part_size < NP;
      AddRandomPair(b_part, a_part, NP))
    { if(!ComputeFixpoint(backup,actual,b_part,a_part,prog,part_size))
	return(false);
    }
  for(int i=0; i<NP; i++) for(int j= 0; j<NP; j++)
    if(b_part[i] == a_part[j])
      { perm[i] = j; break; }
  return(true);
}


bool StateEquivalence(int*backup,int*actual,int*perm, Program&prog)
{ //st = clock();
  bool result;
  stat5++;
  switch(SymmetryOption & 6)
    { case 0: result = StateEquivalence0(backup,actual,perm,prog);break;
      case 2: result = StateEquivalence1(backup,actual,perm,prog);break;
      case 4: result = StateEquivalence2(backup,actual,perm,prog);break;
      case 6: result = StateEquivalence3(backup,actual,perm,prog);break;
    }
  //if(result) stat4+= (clock() - st); else stat5+= (clock() - st);
  return result;
}

bool TranspositionFixes(int *SValues, int d1, int d2, Program& prog)
{ static int       flag = 1;
  static int      *part;
  static int      *perm;
  static int       NP;
  static bool      answer;

  if(flag) { NP   = prog.GetNumberOfProcesses();
	     part = new int [NP];
	     perm = new int [NP];
	     prog.GetModulePartition(part);
	     for(int i = 0; i <NP; i++) perm[i] = i;
	     flag = 0;
	   }

  if(part[d1] != part[d2]) return(false);
  perm[d1] = d2;
  perm[d2] = d1;
  answer = PermutationFixes(SValues, SValues, perm, prog);
  perm[d1] = d1;
  perm[d2] = d2;
  return answer;
}

void FindPartition0(int *SValues, int *partition, Program& prog)
{ for(int i = 0; i<prog.GetNumberOfProcesses(); i++)
    partition[i] = (i<<2) + 3;
}


void FindPartition1(int *SValues, int *partition, Program& prog)
{  static Domains doms(prog);

   prog.GetModulePartition(partition);
   doms.Set(SValues, partition);
   doms.Sort();
   doms.AdjustPartition();
}


void FindPartition2(int *SValues, int *partition, Program& prog)
{ int               i;
  Node<int>      *pNs;
  Node<int>      *pNt;
  Iterator<int>   ItS;
  List<int>        sL;
  int         pNtdata;
  int NumberOfProcesses = prog.GetNumberOfProcesses();

  for(i = 0; i < NumberOfProcesses; i++ )  sL.Append(i);
  
  while(pNt = sL.getHead())      // if the list is empty pNt is set to NULL 
    {                            // and so we exit the loop
      pNtdata = sL.RemoveHead();         
      partition[pNtdata] = (pNtdata << 2) + 3;
                                 // this is  the leader.

      ItS.Adjust(sL);            // Compare everything with pNtdata
      for( pNs = ItS.First(); ItS.More(); pNs=ItS.Next() )
  	if(TranspositionFixes(SValues,pNtdata,pNs->data,prog))
  	  {  partition[pNs->data] = pNtdata << 2;
	     sL.Remove(pNs);     
  	  }
    }
}

void FindPartition(int *SValues, int *partition, Program& prog)
{
   if(SymmetryOption & 1)
    { if(SymmetryOption & 4)
        FindPartition2(SValues, partition, prog);
      else
	FindPartition1(SValues, partition, prog);
    }  
   else FindPartition0(SValues, partition, prog);

}

/* 
                 
SymmetryOption  FindPartition  StateEquivalence
0               0              0
1               1              0
2               0              1
3               1              1            <- Ideal for Client Server
4               0              2
5               2              2
6               0              3
7               2              3

*/
